# -*- coding: utf-8 -*-


# Built-in
import itertools as itt
import warnings


# Common
import numpy as np
from contourpy import contour_generator
import datastock as ds


# specific


# #################################################################
# #################################################################
#               Main
# #################################################################


def get_contours(
    coll=None,
    key=None,
    levels=None,
    res=None,
    npts=None,
    largest=None,
    ref_com=None,
    # return vs store
    returnas=None,
    return_dref=None,
    store=None,
    key_npts=None,
    key_lvls=None,
    key_cont0=None,
    key_cont1=None,
):

    # ----------
    # check

    (
        key, keybs, keym0, levels,
        store, returnas, return_dref,
        key_npts, key_lvls, key_cont0, key_cont1,
    ) = _check(
        coll=coll,
        key=key,
        levels=levels,
        # return vs store
        returnas=returnas,
        return_dref=return_dref,
        store=store,
        key_npts=key_npts,
        key_lvls=key_lvls,
        key_cont0=key_cont0,
        key_cont1=key_cont1,
    )

    # ---------------------
    # prepare (interpolate)

    dsamp = coll.get_sample_mesh(
        key=keym0,
        res=res,
        grid=True,
        store=False,
    )

    # temporary addition of sample
    if ref_com is not None:

        # ref
        lkr_temp = []
        for ii, ss in enumerate(dsamp['x0']['data'].shape):
            kri = f'{key}_refsamptemp{ii}'
            lkr_temp.append(kri)
            coll.add_ref(key=kri, size=ss)

        # data
        lkd_temp = [f'{key}_samptemp0', f'{key}_samptemp1']
        coll.add_data(key=lkd_temp[0], data=dsamp['x0']['data'], ref=lkr_temp)
        coll.add_data(key=lkd_temp[1], data=dsamp['x1']['data'], ref=lkr_temp)
        x0, x1 = lkd_temp
    else:
        x0 = dsamp['x0']['data']
        x1 = dsamp['x1']['data']

    # interpolate
    dinterp = coll.interpolate(
        keys=key,
        ref_key=keybs,
        x0=x0,
        x1=x1,
        grid=False,
        submesh=True,
        ref_com=ref_com,
        store=False,
    )[key]

    # remove temporary data
    if ref_com is not None:
        coll.remove_data(x0, propagate=False)
        coll.remove_data(x1, propagate=False)
        for rr in lkr_temp:
            coll.remove_ref(rr, propagate=False)

        axis = [
            dinterp['ref'].index(lkr_temp[0]),
            dinterp['ref'].index(lkr_temp[1]),
        ]

    else:
        # axis
        axis = [
            ii for ii, rr in enumerate(dinterp['ref'])
            if rr is None
        ]

    # ----------------
    # compute contours

    cont0, cont1 = _get_contours(
        xx0=dsamp['x0']['data'],
        xx1=dsamp['x1']['data'],
        val=dinterp['data'],
        axis=axis,
        levels=levels,
        npts=npts,
        largest=largest,
    )

    # ----------------
    # format output

    dout, dref = _format(
        coll=coll,
        key=key,
        keybs=keybs,
        dinterp=dinterp,
        cont0=cont0,
        cont1=cont1,
        axis=axis,
        keym0=keym0,
        key_npts=key_npts,
        key_lvls=key_lvls,
        key_cont0=key_cont0,
        key_cont1=key_cont1,
    )

    # ----------------
    # store and return

    if store is True:
        _store(
            coll=coll,
            dout=dout,
            dref=dref,
        )

    if returnas is True:
        if return_dref is True:
            return dout, dref
        else:
            return dout


# #################################################################
# #################################################################
#               check
# #################################################################


def _check(
    coll=None,
    key=None,
    levels=None,
    # store vs return
    returnas=None,
    return_dref=None,
    store=None,
    key_npts=None,
    key_lvls=None,
    key_cont0=None,
    key_cont1=None,
):

    # ------
    # key

    dp2d = coll.get_profiles2d()
    key = ds._generic_check._check_var(
        key, 'key',
        types=str,
        allowed=list(dp2d.keys()),
    )

    wm = coll._which_mesh
    wbs = coll._which_bsplines
    keybs = dp2d[key]
    keym0 = coll.dobj[wbs][keybs][wm]
    submesh = coll.dobj[wm][keym0]['submesh']
    if submesh is not None:
        keym0 = submesh

    # ----------
    # store

    store = ds._generic_check._check_var(
        store, 'store',
        types=bool,
        default=False,
    )

    # ----------
    # returnas

    returnas = ds._generic_check._check_var(
        returnas, 'returnas',
        types=bool,
        default=not store,
    )

    # ----------
    # return_dref

    return_dref = ds._generic_check._check_var(
        return_dref, 'return_dref',
        types=bool,
        default=False,
    )

    # -----------------
    # keys for storing

    # key_npts
    if store is True:
        lout = list(coll.dref.keys())
    else:
        lout = []
    key_npts = ds._generic_check._check_var(
        key_npts, 'key_npts',
        types=str,
        default=f'{key}_cont_npts',
        excluded=lout,
    )

    # key_lvls
    lout.append(key_npts)
    key_lvls = ds._generic_check._check_var(
        key_lvls, 'key_lvls',
        types=str,
        default=f'{key}_cont_nlvls',
        excluded=lout,
    )

    # key_cont0
    if store is True:
        lout = list(coll.ddata.keys())
    else:
        lout = []
    key_cont0 = ds._generic_check._check_var(
        key_cont0, 'key_cont0',
        types=str,
        default=f'{key}_cont0',
        excluded=lout,
    )

    # key_cont0
    lout.append(key_cont0)
    key_cont1 = ds._generic_check._check_var(
        key_cont1, 'key_cont1',
        types=str,
        default=f'{key}_cont1',
        excluded=lout,
    )

    return (
        key, keybs, keym0, levels,
        store, returnas, return_dref,
        key_npts, key_lvls, key_cont0, key_cont1,
    )


# #################################################################
# #################################################################
#               compute contours
# #################################################################


def _get_contours(
    xx0=None,
    xx1=None,
    val=None,
    axis=None,
    levels=None,
    npts=None,
    largest=None,
):
    """ Return x0, x1 coordinates of contours (time-dependent)

    For contourpy algorithm, the dimensions shoud be (ny, nx), from meshgrid

    xx0 = (n1, n0)
    xx1 = (n1, n0)
    val = (nt, n1, n0)
    levels = (nlevels,)

    c0 = (nt, nlevels, nmax) array of x0 coordinates
    c1 = (nt, nlevels, nmax) array of x1 coordinates

    The contour coordinates are uniformized to always have the same nb of pts

    DISCLAIMER: not entirely dimension agnostic yet!

    """

    # -------------
    # check inputs

    (
        axis, n0, n1, shape_other, shape_cont, npts, levels, largest,
    ) = _compute_check(
        xx0=xx0,
        xx1=xx1,
        val=val,
        axis=axis,
        levels=levels,
        npts=npts,
        largest=largest,
    )

    # --------
    # prepare

    # compute contours at rknots
    # see https://github.com/matplotlib/matplotlib/blob/main/src/_contour.h

    cont0 = np.full(shape_cont, np.nan)
    cont1 = np.full(shape_cont, np.nan)

    # prepare slices
    sli_v = [slice(None) if ii in axis else 0 for ii in range(val.ndim)]
    sli_c = [slice(None) if ii == axis[0] else 0 for ii in range(cont0.ndim)]

    # prepare indices
    ind_v = [ii for ii in range(val.ndim) if ii not in axis]
    ind_c = [ii for ii in range(cont0.ndim) if ii not in axis]

    # ------------------------
    # loop on slices

    maxnpts = 0
    maxlvl = levels[0]
    for ind in itt.product(*[range(aa) for aa in shape_other]):

        # slice
        for ii, jj in enumerate(ind):
            sli_v[ind_v[ii]] = jj
            sli_c[ind_c[ii]] = jj

        # define map
        contgen = contour_generator(
            x=xx0,
            y=xx1,
            z=val[tuple(sli_v)],
            name='serial',
            corner_mask=None,
            line_type='Separate',
            fill_type=None,
            chunk_size=None,
            chunk_count=None,
            total_chunk_count=None,
            quad_as_tri=True,       # for sub-mesh precision
            # z_interp=<ZInterp.Linear: 1>,
            thread_count=0,
        )

        # loop on levels
        for jj in range(len(levels)):

            # get contour
            no_cont, cj = _get_contours_lvls(
                contgen=contgen,
                level=levels[jj],
                largest=largest,
            )

            # if contour was found
            if no_cont is False:

                # maxnpts
                if cj.shape[0] > maxnpts:
                    maxnpts = cj.shape[0]
                    maxlvl = levels[jj]

                # slice
                sli_c[axis[1]] = jj

                # interpolate on desired nb of points
                cont0[tuple(sli_c)] = np.interp(
                    np.linspace(0, cj.shape[0], npts),
                    np.arange(0, cj.shape[0]),
                    cj[:, 0],
                )
                cont1[tuple(sli_c)] = np.interp(
                    np.linspace(0, cj.shape[0], npts),
                    np.arange(0, cj.shape[0]),
                    cj[:, 1],
                )

    # -----------------
    # Warning + return

    msg = f"Contour computing: npts {npts} vs {maxnpts} (level {maxlvl})"
    if maxnpts > npts:
        warnings.warn(msg)
    else:
        print(msg)

    return cont0, cont1


def _get_contours_lvls(
    contgen=None,
    level=None,
    largest=None,
):
    # compute concatenated contour
    no_cont = False
    cj = contgen.lines(level)

    c0 = (
        isinstance(cj, list)
        and all([
            isinstance(cjj, np.ndarray)
            and cjj.ndim == 2
            and cjj.shape[1] == 2
            for cjj in cj
        ])
    )
    if not c0:
        msg = f"Wrong output from contourpy!\n{cj}"
        raise Exception(msg)

    # if one or several contours exist
    if len(cj) > 0:
        cj = [
            cc[np.all(np.isfinite(cc), axis=1), :]
            for cc in cj
            if np.sum(np.all(np.isfinite(cc), axis=1)) >= 3
        ]

        if len(cj) == 0:
            no_cont = True
        elif len(cj) == 1:
            cj = cj[0]
        elif len(cj) > 1:
            if largest:
                nj = [
                    0.5*np.abs(np.sum(
                        (cc[1:, 0] + cc[:-1, 0])
                        *(cc[1:, 1] - cc[:-1, 1])
                    ))
                    for cc in cj
                ]
                cj = cj[np.argmax(nj)]
            else:
                ij = np.cumsum([cc.shape[0] for cc in cj])
                cj = np.concatenate(cj, axis=0)
                cj = np.insert(cj, ij, np.nan, axis=0)

        elif np.sum(np.all(~np.isnan(cj), axis=1)) < 3:
            no_cont = True
    else:
        no_cont = True

    return no_cont, cj


def _compute_check(
    xx0=None,
    xx1=None,
    val=None,
    axis=None,
    levels=None,
    npts=None,
    largest=None,
):

    # ---------------------
    # check axis and shapes

    if axis is None:
        axis = [
            ii for ii, ss in enumerate(val.shape)
            if ss in xx0.shape
        ]
    assert xx0.ndim == 2
    assert xx0.shape == xx1.shape
    assert len(axis) == 2

    # ---------------------
    # check axis and shapes

    npts = ds._generic_check._check_var(
        npts, 'npts',
        types=int,
        default=100,
        sign='>0',
    )

    # ----------
    # levels

    levels = ds._generic_check._check_flat1darray(
        levels, 'levels',
        dtype=float,
    )

    # ---------------------
    # check axis and shapes

    n0, n1 = xx0.shape
    shape_other = [
        ss for ii, ss in enumerate(val.shape)
        if ii not in axis
    ]
    shape_cont = tuple([
        npts if ii == axis[0]
        else (levels.size if ii == axis[1] else ss)
        for ii, ss in enumerate(val.shape)
    ])

    # ----------
    # largest

    largest = ds._generic_check._check_var(
        largest, 'largest',
        types=bool,
        default=True,
    )

    return axis, n0, n1, shape_other, shape_cont, npts, levels, largest


# #################################################################
# #################################################################
#               format
# #################################################################


def _format(
    coll=None,
    key=None,
    keybs=None,
    dinterp=None,
    cont0=None,
    cont1=None,
    axis=None,
    keym0=None,
    key_npts=None,
    key_lvls=None,
    key_cont0=None,
    key_cont1=None,
):

    npts = cont0.shape[axis[0]]
    nlvls = cont0.shape[axis[1]]

    # --------
    # ref

    ref = tuple([
        key_npts if ii == axis[0]
        else (key_lvls if ii == axis[1] else rr)
        for ii, rr in enumerate(dinterp['ref'])
    ])

    # -------
    # units

    wm = coll._which_mesh
    knots0, knots1 = coll.dobj[wm][keym0]['knots']

    lf = ['units', 'dim', 'quant', 'name']
    dd0 = {k0: coll.ddata[knots0][k0] for k0 in lf}
    dd1 = {k0: coll.ddata[knots1][k0] for k0 in lf}

    # -------------
    # populate dict

    dref = {
        'npts': {'key': key_npts, 'size': npts},
        'levels': {'key': key_lvls, 'size': nlvls},
    }

    dout = {
        'cont0': {
            'key': key_cont0,
            'data': cont0,
            'ref': ref,
            **dd0,
        },
        'cont1': {
            'key': key_cont1,
            'data': cont1,
            'ref': ref,
            **dd1,
        },
    }

    return dout, dref


# #################################################################
# #################################################################
#               store
# #################################################################


def _store(
    coll=None,
    dout=None,
    dref=None,
):

    # --------
    # add ref

    for k0, v0 in dref.items():
        coll.add_ref(**v0)

    # --------
    # add data

    for k0, v0 in dout.items():
        coll.add_data(**v0)


# #############################################################################
# #############################################################################
#                   Polygon simplification
# #############################################################################


# def _simplify_polygon(pR=None, pZ=None, res=None):
    # """ Use convex hull with a constraint on the maximum discrepancy """

    # # ----------
    # # preliminary 1: check there is non redundant point

    # dp = np.sqrt((pR[1:] - pR[:-1])**2 + (pZ[1:] - pZ[:-1])**2)
    # ind = (dp > 1.e-6).nonzero()[0]
    # pR = pR[ind]
    # pZ = pZ[ind]

    # # check new poly is closed
    # if (pR[0] != pR[-1]) or (pZ[0] != pZ[-1]):
        # pR = np.append(pR, pR[0])
        # pZ = np.append(pZ, pZ[0])

    # # check it is counter-clockwise
    # clock = np.nansum((pR[1:] - pR[:-1]) * (pZ[1:] + pZ[:-1]))
    # if clock > 0:
        # pR = pR[::-1]
        # pZ = pZ[::-1]

    # # threshold = diagonal of resolution + 10%
    # thresh = res * np.sqrt(2) * 1.1

    # # ----------
    # # preliminary 2: get convex hull and copy

    # poly = np.array([pR, pZ]).T
    # iconv = ConvexHull(poly, incremental=False).vertices

    # # close convex hull to iterate on edges
    # pR_conv = np.append(pR[iconv], pR[iconv[0]])
    # pZ_conv = np.append(pZ[iconv], pZ[iconv[0]])

    # # copy to create new polygon that will serve as buffer
    # pR_bis, pZ_bis = np.copy(pR), np.copy(pZ)

    # # -------------------------
    # # loop on convex hull edges

    # for ii in range(pR_conv.size - 1):

        # pR1, pR2 = pR_conv[ii], pR_conv[ii+1]
        # pZ1, pZ2 = pZ_conv[ii], pZ_conv[ii+1]
        # i0 = np.argmin(np.hypot(pR_bis - pR1, pZ_bis - pZ1))

        # # make sure it starts from p1
        # pR_bis = np.append(pR_bis[i0:], pR_bis[:i0])
        # pZ_bis = np.append(pZ_bis[i0:], pZ_bis[:i0])

        # # get indices of closest points to p1, p2
        # i1 = np.argmin(np.hypot(pR_bis - pR1, pZ_bis - pZ1))
        # i2 = np.argmin(np.hypot(pR_bis - pR2, pZ_bis - pZ2))

        # # get corresponding indices of poly points to be included
        # if i2 == i1 + 1:
            # itemp = [i1, i2]

        # else:
            # # several points in-between
            # # => check they are all within distance before exclusing them

            # # get unit vector of segment
            # norm12 = np.hypot(pR2 - pR1, pZ2 - pZ1)
            # u12R = (pR2 - pR1) / norm12
            # u12Z = (pZ2 - pZ1) / norm12

            # # get points standing between p1 nd p2
            # lpR = pR_bis[i1 + 1:i2]
            # lpZ = pZ_bis[i1 + 1:i2]

            # # indices of points standing too far from edge (use cross-product)
            # iout = np.abs(u12R*(lpZ - pZ1) - u12Z*(lpR - pR1)) > thresh

            # # if any pts too far => include all pts
            # if np.any(iout):
                # itemp = np.arange(i1, i2 + 1)
            # else:
                # itemp = [i1, i2]

        # # build pts_in
        # pR_in = pR_bis[itemp]
        # pZ_in = pZ_bis[itemp]

        # # concatenate to add to new polygon
        # pR_bis = np.append(pR_in, pR_bis[i2 + 1:])
        # pZ_bis = np.append(pZ_in, pZ_bis[i2 + 1:])

    # # check new poly is closed
    # if (pR_bis[0] != pR_bis[-1]) or (pZ_bis[0] != pZ_bis[-1]):
        # pR_bis = np.append(pR_bis, pR_bis[0])
        # pZ_bis = np.append(pZ_bis, pZ_bis[0])

    # return pR_bis, pZ_bis


# #############################################################################
# #############################################################################
#                   radius2d special points handling
# #############################################################################


# def radius2d_special_points(
    # coll=None,
    # key=None,
    # keym0=None,
    # res=None,
# ):

    # keybs = coll.ddata[key]['bsplines']
    # keym = coll.dobj['bsplines'][keybs]['mesh']
    # mtype = coll.dobj[coll._which_mesh][keym]['type']
    # assert mtype in ['rect', 'tri']

    # # get map sampling
    # RR, ZZ = coll.get_sample_mesh(
        # key=keym,
        # res=res,
        # grid=True,
    # )

    # # get map
    # val, t, _ = coll.interpolate_profile2d(
        # key=key,
        # R=RR,
        # Z=ZZ,
        # grid=False,
        # imshow=True,        # for contour
    # )

    # # get min max values
    # rmin = np.nanmin(val)
    # rmax = np.nanmax(val)

    # # get contour of 0
    # cR, cZ = _get_contours(
        # RR=RR,
        # ZZ=ZZ,
        # val=val,
        # levels=[rmin + 0.05*(rmax-rmin)],
    # )

    # # dref
    # ref_O = f'{keym0}-pts-O-n'
    # dref = {
        # ref_O: {'size': 1},
    # }

    # # get barycenter
    # if val.ndim == 3:
        # assert cR.shape[1] == 1
        # ax_R = np.nanmean(cR[:, 0, :], axis=-1)[:, None]
        # ax_Z = np.nanmean(cZ[:, 0, :], axis=-1)[:, None]
        # reft = coll.ddata[key]['ref'][0]
        # ref = (reft, ref_O)
    # else:
        # ax_R = np.r_[np.nanmean(cR)]
        # ax_Z = np.r_[np.nanmean(cZ)]
        # ref = (ref_O,)

    # kR = f'{keym0}-pts-O-R'
    # kZ = f'{keym0}-pts-O-Z'
    # ddata = {
        # kR: {
            # 'ref': ref,
            # 'data': ax_R,
            # 'dim': 'distance',
            # 'quant': 'R',
            # 'name': 'O-points_R',
            # 'units': 'm',
        # },
        # kZ: {
            # 'ref': ref,
            # 'data': ax_Z,
            # 'dim': 'distance',
            # 'quant': 'Z',
            # 'name': 'O-points_Z',
            # 'units': 'm',
        # },
    # }

    # return dref, ddata, kR, kZ


# #############################################################################
# #############################################################################
#                   angle2d discontinuity handling
# #############################################################################


# def angle2d_zone(
    # coll=None,
    # key=None,
    # keyrad2d=None,
    # key_ptsO=None,
    # res=None,
    # keym0=None,
# ):

    # keybs = coll.ddata[key]['bsplines']
    # keym = coll.dobj['bsplines'][keybs]['mesh']
    # mtype = coll.dobj[coll._which_mesh][keym]['type']
    # assert mtype in ['rect', 'tri']

    # # --------------
    # # prepare

    # hastime, hasvect, reft, keyt = coll.get_time(key=key)[:4]
    # if hastime:
        # nt = coll.dref[reft]['size']
    # else:
        # msg = (
            # "Non time-dependent angle2d not implemented yet\n"
            # "=> ping @Didou09 on Github to open an issue"
        # )
        # raise NotImplementedError(msg)

    # if res is None:
        # res = _get_sample_mesh_res(
            # coll=coll,
            # keym=keym,
            # mtype=mtype,
        # )

    # # get map sampling
    # RR, ZZ = coll.get_sample_mesh(
        # key=keym,
        # res=res/2.,
        # grid=True,
        # imshow=True,    # for contour
    # )

    # # get map
    # val, t, _ = coll.interpolate_profile2d(
        # key=key,
        # R=RR,
        # Z=ZZ,
        # grid=False,
        # azone=False,
    # )
    # val[np.isnan(val)] = 0.
    # amin = np.nanmin(val)
    # amax = np.nanmax(val)

    # # get contours of absolute value
    # cRmin, cZmin = _get_contours(
        # RR=RR,
        # ZZ=ZZ,
        # val=val,
        # levels=[amin + 0.10*(amax - amin)],
        # largest=True,
        # uniform=True,
    # )
    # cRmax, cZmax = _get_contours(
        # RR=RR,
        # ZZ=ZZ,
        # val=val,
        # levels=[amax - 0.10*(amax - amin)],
        # largest=True,
        # uniform=True,
    # )

    # cRmin, cZmin = cRmin[:, 0, :], cZmin[:, 0, :]
    # cRmax, cZmax = cRmax[:, 0, :], cZmax[:, 0, :]

    # rmin = np.full(cRmin.shape, np.nan)
    # rmax = np.full(cRmax.shape, np.nan)

    # # get points inside contour
    # for ii in range(nt):
        # rmin[ii, :], _, _ = coll.interpolate_profile2d(
            # key=keyrad2d,
            # R=cRmin[ii, :],
            # Z=cZmin[ii, :],
            # grid=False,
            # indt=ii,
        # )
        # rmax[ii, :], _, _ = coll.interpolate_profile2d(
            # key=keyrad2d,
            # R=cRmax[ii, :],
            # Z=cZmax[ii, :],
            # grid=False,
            # indt=ii,
        # )

    # # get magnetic axis
    # kR, kZ = key_ptsO
    # axR = coll.ddata[kR]['data']
    # axZ = coll.ddata[kZ]['data']
    # assert coll.ddata[kR]['ref'][0] == coll.ddata[key]['ref'][0]

    # start_min = np.nanargmin(rmin, axis=-1)
    # start_max = np.nanargmin(rmax, axis=-1)

    # # re-order from start_min, start_max
    # lpR, lpZ = [], []
    # for ii in range(rmin.shape[0]):
        # imin = np.r_[
            # np.arange(start_min[ii], rmin.shape[1]),
            # np.arange(0, start_min[ii]),
        # ]

        # cRmin[ii] = cRmin[ii, imin]
        # cZmin[ii] = cZmin[ii, imin]
        # rmin[ii] = rmin[ii, imin]
        # # check it is counter-clockwise
        # clock = np.nansum(
            # (cRmin[ii, 1:] - cRmin[ii, :-1])
            # *(cZmin[ii, 1:] + cZmin[ii, :-1])
        # )
        # if clock > 0:
            # cRmin[ii, :] = cRmin[ii, ::-1]
            # cZmin[ii, :] = cZmin[ii, ::-1]
            # rmin[ii, :] = rmin[ii, ::-1]

        # imax = np.r_[
            # np.arange(start_max[ii], rmax.shape[1]),
            # np.arange(0, start_max[ii])
        # ]
        # cRmax[ii] = cRmax[ii, imax]
        # cZmax[ii] = cZmax[ii, imax]
        # rmax[ii] = rmax[ii, imax]
        # # check it is clockwise
        # clock = np.nansum(
            # (cRmax[ii, 1:] - cRmax[ii, :-1])
            # *(cZmax[ii, 1:] + cZmax[ii, :-1])
        # )
        # if clock < 0:
            # cRmax[ii, :] = cRmax[ii, ::-1]
            # cZmax[ii, :] = cZmax[ii, ::-1]
            # rmax[ii, :] = rmax[ii, ::-1]

        # # i0
        # dr = np.diff(rmin[ii, :])
        # i0 = (np.isnan(dr) | (dr < 0)).nonzero()[0][0]
        # # rmin[ii, i0-1:] = np.nan
        # dr = np.diff(rmax[ii, :])
        # i1 = (np.isnan(dr) | (dr < 0)).nonzero()[0][0]
        # # rmax[ii, i1-1:] = np.nan

        # # polygon
        # pR = np.r_[axR[ii], cRmin[ii, :i0-1], cRmax[ii, :i1-1][::-1]]
        # pZ = np.r_[axZ[ii], cZmin[ii, :i0-1], cZmax[ii, :i1-1][::-1]]

        # pR, pZ = _simplify_polygon(pR=pR, pZ=pZ, res=res)

        # lpR.append(pR)
        # lpZ.append(pZ)

    # # Ajust sizes
    # nb = np.array([pR.size for pR in lpR])

    # #
    # nmax = np.max(nb)
    # pR = np.full((nt, nmax), np.nan)
    # pZ = np.full((nt, nmax), np.nan)

    # for ii in range(nt):
        # pR[ii, :] = np.interp(
            # np.linspace(0, nb[ii], nmax),
            # np.arange(0, nb[ii]),
            # lpR[ii],
        # )
        # pZ[ii, :] = np.interp(
            # np.linspace(0, nb[ii], nmax),
            # np.arange(0, nb[ii]),
            # lpZ[ii],
        # )

    # # ----------------
    # # prepare output dict

    # # ref
    # kref = f'{keym0}-azone-npt'
    # dref = {
        # kref: {'size': nmax}
    # }

    # # data
    # kR = f'{keym0}-azone-R'
    # kZ = f'{keym0}-azone-Z'
    # ddata = {
        # kR: {
            # 'data': pR,
            # 'ref': (reft, kref),
            # 'units': 'm',
            # 'dim': 'distance',
            # 'quant': 'R',
            # 'name': None,
        # },
        # kZ: {
            # 'data': pZ,
            # 'ref': (reft, kref),
            # 'units': 'm',
            # 'dim': 'distance',
            # 'quant': 'R',
            # 'name': None,
        # },
    # }

    # return dref, ddata, kR, kZ


# def angle2d_inzone(
    # coll=None,
    # keym0=None,
    # keya2d=None,
    # R=None,
    # Z=None,
    # t=None,
    # indt=None,
# ):


    # # ------------
    # # prepare points

    # if R.ndim == 1:
        # shape0 = None
        # pts = np.array([R, Z]).T
    # else:
        # shape0 = R.shape
        # pts = np.array([R.ravel(), Z.ravel()]).T

    # # ------------
    # # prepare path

    # kazR, kazZ = coll.dobj[coll._which_mesh][keym0]['azone']
    # pR = coll.ddata[kazR]['data']
    # pZ = coll.ddata[kazZ]['data']

    # hastime, hasvect, reft, keyt, tnew, dind = coll.get_time(
        # key=kazR,
        # t=t,
        # indt=indt,
    # )

    # # ------------
    # # test points

    # if hastime:
        # if dind is None:
            # nt = coll.dref[reft]['size']
            # ind = np.zeros((nt, R.size), dtype=bool)
            # for ii in range(nt):
                # path = Path(np.array([pR[ii, :], pZ[ii, :]]).T)
                # ind[ii, :] = path.contains_points(pts)
        # else:
            # import pdb; pdb.set_trace()     # DB
            # raise NotImplementedError()
            # # TBC / TBF
            # nt = None
            # ind = np.zeros((nt, R.size), dtype=bool)
            # for ii in range(nt):
                # path = Path(np.array([pR[ii, :], pZ[ii, :]]).T)
                # ind[ii, :] = path.contains_points(pts)

    # else:
        # path = Path(np.array([pR, pZ]).T)
        # ind = path.contains_points(pts)

    # # -------------------------
    # # fromat output and return

    # if shape0 is not None:
        # if hastime:
            # ind = ind.reshape(tuple(np.r_[nt, shape0]))
        # else:
            # ind = ind.reshape(shape0)

    # return ind



# -*- coding: utf-8 -*-


# Built-in
import itertools as itt
import warnings

# Common
import numpy as np
import datastock as ds


# tofu
# from . import _class02_checks as _checks
from . import _utils_bsplines


# ################################################################
# ################################################################
#                                       Main
# ################################################################


def interpolate(
    coll=None,
    # interpolation base
    keys=None,
    ref_key=None,
    # interpolation pts
    x0=None,
    x1=None,
    grid=None,
    res=None,
    mode=None,
    submesh=None,
    # domain limitation
    domain=None,
    # common ref
    ref_com=None,
    ref_vector_strategy=None,
    # bsplines-specific
    details=None,
    indbs_tf=None,
    # rect-specific
    crop=None,
    # parameters
    deg=None,
    deriv=None,
    val_out=None,
    log_log=None,
    nan0=None,
    # store vs return
    returnas=None,
    return_params=None,
    store=None,
    store_keys=None,
    inplace=None,
    # debug or unit tests
    debug=None,
):
    """ Interpolate at desired points on desired data

    Interpolate quantities (keys) on coordinates (ref_keys)
    All provided keys should share the same refs
    They should have dimension 2 or less

    ref_keys should be a list of monotonous data keys to be used as coordinates
    It should have one element per dimension in refs

    The interpolation points are provided as np.ndarrays in each dimension
    They should all have the same shape except if grid = True

    deg is the degree of the interpolation

    It is an interpolation, not a smoothing, so it will pass through all points
    Uses scpinterp.InterpolatedUnivariateSpline() for 1d

    """

    # -------------
    # check inputs

    if debug is None:
        debug = False

    # keys
    (
        isbs, keys, ref_key,
        daxis, dunits, units_ref,
        details, ktemp,
    ) = _check_keys(
        coll=coll,
        keys=keys,
        ref_key=ref_key,
        only1d=False,
        details=details,
    )

    # -----------
    # isbs

    x0_str = None
    if isbs:

        # -----------------
        # prepare bsplines

        keybs, keym, mtype, ref_key = _prepare_bsplines(
            coll=coll,
            ref_key=ref_key,
        )

        if mtype == 'tri' and ref_com is not None:
            msg = (
                "Arg ref_com cannot be used with a triangular mesh!"
            )
            raise NotImplementedError(msg)

        # ------------------------------
        # check bsplines-specific params

        (
            val_out, crop, cropbs, indbs_tf, nbs, submesh,
        ) = _check_params_bsplines(
            coll=coll,
            details=details,
            val_out=val_out,
            crop=crop,
            mtype=mtype,
            keybs=keybs,
            keym=keym,
            indbs_tf=indbs_tf,
            submesh=submesh,
        )

        if details is True:
            ref_com, domain = None, None

        # ---------------
        # no x0 => mesh

        wm = coll._which_mesh
        if x0 is None:
            if submesh is True:
                km = coll.dobj[wm][keym]['submesh']
            else:
                km = keym

            out = coll.get_sample_mesh(
                key=km,
                res=res,
                grid=True,
                mode=mode,
                x0=None,
                x1=None,
                Dx0=None,
                Dx1=None,
                imshow=False,
            )

            x0 = out['x0']['data']
            x1 = out.get('x1', {}).get('data')

        # ---------------
        # submesh

        if submesh is True:

            # submesh
            kd0 = coll.dobj[wm][keym]['subkey']
            kbs0 = coll.dobj[wm][keym]['subbs']

            # interpolate
            dout_temp, dparams_temp = coll.interpolate(
                # interpolation base
                keys=kd0,
                ref_key=kbs0,
                # interpolation pts
                x0=x0,
                x1=x1,
                grid=grid,
                submesh=False,
                # domain
                domain=domain,
                # common ref
                ref_com=None,
                # bsplines-specific
                details=False,
                indbs_tf=None,
                # rect-specific
                crop=crop,
                return_params=True,
                store=False,
            )

            # ------------------------------------------
            # handle ref_com, add ref and data if needed

            ref_com = _submesh_ref_com(
                coll=coll,
                kd0=kd0,
                keys=keys,
                ref_com=ref_com,
                # coordinates
                x0=x0,
            )

            # temporary storing for ref handling
            lr_add, ld_add = _submesh_addtemp(
                coll=coll,
                kd0=kd0,
                # interp resut
                dout_temp=dout_temp,
                # coordinates
                x0=x0,
            )

            # --------
            # substitue x0, x1

            x0_str = isinstance(x0, str)
            # if ref_com is None:
                # x0 = dout_temp[kd0[0]]['data']
                # if len(kd0) > 1:
                    # x1 = dout_temp[kd0[1]]['data']
            # else:
            x0 = dout_temp[kd0[0]]['key']
            if len(kd0) > 1:
                x1 = dout_temp[kd0[1]]['key']

            if len(kd0) == 1:
                x1 = None

        else:
            lr_add, ld_add = None, None

    # ---------------
    # prepare x0, x1

    # params
    (
        deg, deriv,
        kx0, kx1, x0, x1, refx, dref_com,
        ddata, dout, dsh_other, sli_c, sli_x, sli_v,
        log_log, nan0, grid, ndim, xunique,
        returnas, return_params, store, inplace,
    ) = ds._class1_interpolate._check(
        coll=coll,
        # interpolation base
        keys=keys,
        ref_key=ref_key,      # ddata keys
        # interpolation pts
        x0=x0,
        x1=x1,
        # useful for shapes
        daxis=daxis,
        dunits=dunits,
        domain=domain,
        # common ref
        ref_com=ref_com,
        # ref_vector_strategy=ref_vector_strategy,
        # parameters
        grid=grid,
        deg=None,
        deriv=deriv,
        log_log=None,
        nan0=nan0,
        # return vs store
        returnas=returnas,
        return_params=return_params,
        store=store,
        inplace=inplace,
        x0_str=x0_str,
    )

    # ---------------
    # interpolate

    if isbs:

        _interp(
            coll=coll,
            keybs=keybs,
            keys=keys,
            x0=x0,
            x1=x1,
            val_out=val_out,
            deriv=deriv,
            indbs_tf=indbs_tf,
            crop=crop,
            cropbs=cropbs,
            details=details,
            dout=dout,
            mtype=mtype,
            ddata=ddata,
            daxis=daxis,
            sli_c=sli_c,
            sli_x=sli_x,
            sli_v=sli_v,
            dsh_other=dsh_other,
            dref_com=dref_com,
            lr_add=lr_add,
            ld_add=ld_add,
            nan0=nan0,
        )

    else:

        ds._class1_interpolate._interp(
            coll=coll,
            keys=keys,
            ref_key=ref_key,
            x0=x0,
            x1=x1,
            ndim=ndim,
            log_log=log_log,
            dout=dout,
            ddata=ddata,
            dsh_other=dsh_other,
            daxis=daxis,
            deg=deg,
            deriv=deriv,
            dref_com=dref_com,
            sli_x=sli_x,
            sli_c=sli_c,
            sli_v=sli_v,
            nan0=nan0,
        )

    # ------------------------------
    # cleanup if details

    if details is True:
        coll.remove_data(ktemp)

    # ------------------------------
    # adjust data and ref if xunique

    if xunique:
        # try:
        ds._class1_interpolate._xunique(dout)
        # except Exception as err:
        #     msg = (
        #         err.args[0]
        #         + "\n\n"
        #         f"keys = {keys}\n"
        #         f"ref_key = {ref_key}\n"
        #         f"x0 = {x0}\n"
        #         f"x1 = {x1}\n"
        #     )
        #     err.args = (msg,)
        #     raise err
            
            
    # ----------
    # store

    if store is True:
        coll2 = ds._class1_interpolate._store(
            coll=coll,
            dout=dout,
            inplace=inplace,
            store_keys=store_keys,
        )

    # -------
    # return

    if returnas is object:
        return coll2

    elif returnas is dict:

        if return_params is True:
            dparam = {
                'keys': keys,
                'keybs': keybs,
                'ref_key': ref_key,
                'deg': deg,
                'deriv': deriv,
                'log_log': log_log,
                'kx0': kx0,
                'kx1': kx1,
                'x0': x0,
                'x1': x1,
                'grid': grid,
                'refx': refx,
                'ref_com': ref_com,
                'dref_com': dref_com,
                'daxis': daxis,
                'dsh_other': dsh_other,
                'domain': domain,
                'details': details,
                'crop': crop,
                'cropbs': cropbs,
                'indbs_tf': indbs_tf,
                'submesh': submesh,
                'subbs': kbs0 if submesh else None,
            }

            # debug
            if debug is True:
                dparam['x0.shape'] = x0.shape

            return dout, dparam
        else:
            return dout


# ####################################
# ####################################
#       check
# ####################################


def _check_keys(
    coll=None,
    keys=None,
    ref_key=None,
    only1d=None,
    details=None,
):

    # -------------
    # details

    details = ds._generic_check._check_var(
        details, 'details',
        types=bool,
        default=False,
    )

    # -------------
    # only1d

    only1d = ds._generic_check._check_var(
        only1d, 'only1d',
        types=bool,
        default=True,
    )

    maxd = 1 if only1d else 2

    # ---------------
    # details is True

    wbs = coll._which_bsplines
    lbs = list(coll.dobj[wbs].keys())

    ktemp = None
    if details is True:
        ref_key = ds._generic_check._check_var(
            ref_key, 'ref_key',
            types=str,
            allowed=lbs,
        )

        # add temporary data
        ktemp = f'{ref_key}_details'
        coll.add_data(
            key=ktemp,
            data=np.ones(coll.dobj[wbs][ref_key]['shape'], dtype=float),
            ref=ref_key,
        )
        keys = ktemp

    # ---------------
    # keys vs ref_key

    # ref_key
    dbs = coll.get_dict_bsplines()[0]
    if ref_key is not None:

        # basic checks
        if isinstance(ref_key, str):
            ref_key = (ref_key,)

        lref = list(coll.dref.keys())
        ldata = list(coll.ddata.keys())

        ref_key = list(ds._generic_check._check_var_iter(
            ref_key, 'ref_key',
            types=(list, tuple),
            types_iter=str,
            allowed=lref + ldata + lbs,
        ))

        lc = [
            all([rr in lref + ldata for rr in ref_key]),
            all([rr in lbs for rr in ref_key]),
        ]
        if np.sum(lc) != 1:
            msg = (
                "Arg ref_key must refer to (ref or vector) xor bsplines!\n"
                f"Provided: {ref_key}"
            )
            raise Exception(msg)

        # check vs maxd
        if (lc[0] and len(ref_key) > maxd) or (lc[1] and len(ref_key) != 1):
            msg = (
                f"Arg ref_key shall not more than {maxd} ref!\n"
                "And can only contain one single bsplines!\n"
                f"Provided: {ref_key}"
            )
            raise Exception(msg)

        # bs vs refs
        if lc[1]:
            ref_key = ref_key[0]
            lok_bs = [
                k0 for k0, v0 in coll.ddata.items()
                if v0[wbs] is not None and ref_key in v0[wbs]
            ]
            lok_nobs = []

        else:

            # check vs valid vectors
            for ii, rr in enumerate(ref_key):
                if rr in lref:
                    kwd = {'ref': rr}
                else:
                    kwd = {'key': rr}
                hasref, hasvect, ref, ref_key[ii] = coll.get_ref_vector(**kwd)[:4]

                if not (hasref and hasvect):
                    msg = (
                        f"Provided ref_key[{ii}] not a valid ref or ref vector!\n"
                        f"Provided: {rr}"
                    )
                    raise Exception(msg)

            if not (hasref and hasvect):
                msg = (
                    "Provided ref_key not a valid ref or ref vector!\n"
                    "Provided: {ref_key}"
                )
                raise Exception(msg)

            lok_nobs = [
                k0 for k0, v0 in coll.ddata.items()
                if all([coll.ddata[rr]['ref'][0] in v0['ref'] for rr in ref_key])
            ]
            lok_bs = []

        if keys is None:
            keys = lok_nobs + lok_bs

    else:
        # binning only for non-bsplines or 1d bsplines
        lok_nobs = [k0 for k0, v0 in coll.ddata.items() if k0 not in dbs.keys()]
        lok_bs = [
            k0 for k0, v0 in coll.ddata.items()
            if (
                k0 in dbs.keys()
                and any([len(v1) <= maxd for v1 in dbs[k0].values()])
            )
        ]

    # ---------
    # keys

    if isinstance(keys, str):
        keys = [keys]

    keys = ds._generic_check._check_var_iter(
        keys, 'keys',
        types_iter=str,
        types=(list, tuple),
        allowed=lok_nobs + lok_bs,
    )

    libs = [
        all([k0 in lok_bs for k0 in keys]),
        all([k0 not in lok_bs for k0 in keys]),
    ]
    if np.sum(libs) != 1:
        msg = (
            "Either all keys must refer to bsplines or to non-bsplines!\n"
            f"Provided: {keys}"
        )
        raise Exception(msg)

    isbs = libs[0]

    # ------------
    # ref_key

    if isbs:

        # check ref_key
        lbs = [
            [
                bs for bs in coll.ddata[k0][wbs]
                if len(coll.dobj[wbs][bs]['ref']) <= maxd
            ]
            for k0 in keys
            if any([
                len(coll.dobj[wbs][bs]['ref']) <= maxd
                for bs in coll.ddata[k0][wbs]
            ])
        ]
        lbsu = sorted(set(itt.chain.from_iterable(lbs)))
        lbsok = [
            bs for bs in lbsu
            if all([bs in bb for bb in lbs])
        ]

        # ref_key
        ref_key = ds._generic_check._check_var(
            ref_key, 'ref_key',
            types=str,
            allowed=lbsok,
        )

        # daxis
        daxis = {
            k0: [
                coll.ddata[k0]['ref'].index(rr)
                for rr in coll.dobj[wbs][ref_key]['ref']
            ]
            for k0 in keys
        }

        # units_ref
        wbs = coll._which_bsplines
        units_ref = coll.ddata[coll.dobj[wbs][ref_key]['apex'][0]]['units']

    # ref_key
    else:
        if ref_key is None:
            hasref, ref, ref_key, val, dkeys = coll.get_ref_vector_common(
                keys=keys,
            )
            if ref_key is None:
                msg = f"No matching ref vector found for {keys}"
                raise Exception(msg)
            ref_key = (ref_key,)

        # daxis
        daxis = {
            k0: [
                coll.ddata[k0]['ref'].index(coll.ddata[rr]['ref'][0])
                for rr in ref_key
            ]
            for k0 in keys
        }

        # units_ref
        units_ref = [coll.ddata[rr]['units'] for rr in ref_key]

    # dunits
    dunits = {k0: coll.ddata[k0]['units'] for k0 in keys}

    return isbs, keys, ref_key, daxis, dunits, units_ref, details, ktemp


def _check_params_bsplines(
    coll=None,
    details=None,
    val_out=None,
    crop=None,
    mtype=None,
    keybs=None,
    keym=None,
    indbs_tf=None,
    submesh=None,
):

    # -------------
    # val_out

    val_out = ds._generic_check._check_var(
        val_out, 'val_out',
        default=np.nan,
        allowed=[0., np.nan, False],
    )

    # ---------------
    # cropping ?

    wbs = coll._which_bsplines
    cropbs = coll.dobj[wbs][keybs]['crop']
    if cropbs is None:
        cropbs = False
    if cropbs is not False:
        cropbs = coll.ddata[cropbs]['data']
        nbs = cropbs.sum()
    else:
        nbs = np.prod(coll.dobj[wbs][keybs]['shape'])

    # -------------
    # crop

    lok = [False]
    if mtype == 'rect' and cropbs is not False:
        lok.append(True)
    else:
        crop = False

    crop = ds._generic_check._check_var(
        crop, 'crop',
        types=bool,
        default=lok[-1],
        allowed=lok,
    )

    # ---------------
    # indbs_tf

    if details is True:

        if mtype == 'rect':
            returnas = 'tuple-flat'
            # returnas = 'array-flat'
        else:
            returnas = int

        # compute validated indbs array with appropriate form
        indbs_tf = coll.select_ind(
            key=keybs,
            returnas=returnas,
            ind=indbs_tf,
            crop=crop,
        )

        if isinstance(indbs_tf, tuple):
            nbs = indbs_tf[0].size
        else:
            nbs = indbs_tf.size

    # ---------
    # submesh

    submesh = ds._generic_check._check_var(
        submesh, 'submesh',
        types=bool,
        default=False,
    )
    if coll.dobj[coll._which_mesh][keym]['subkey'] is None:
        submesh = False

    return val_out, crop, cropbs, indbs_tf, nbs, submesh


# ################################################################
# ################################################################
#               Handle ref_com for submesh
# ################################################################


def _submesh_ref_com(
    coll=None,
    kd0=None,
    keys=None,
    ref_com=None,
    # coordinates
    x0=None,
):

    # ---------------------
    # find possible matches

    ref0 = coll.ddata[kd0[0]]['ref']
    lrcom = [
        (rr, ref0.index(rr))
        for ii, rr in enumerate(ref0)
        if rr in list(itt.chain.from_iterable([
            coll.ddata[kk]['ref'] for kk in keys
        ]))
        and ii in [0, len(ref0) - 1]
    ]

    # ----------
    # unused options

    if ref_com is None:
        if len(lrcom) > 0:
            msg = (
                f"\nPossible common ref for data {keys} and subkey '{kd0}':\n"
                + "\n".join([f"\t- {rr}" for rr in lrcom])
                + "\nIf you wish to use one, specify with ref_com=..."
            )
            warnings.warn(msg)

    else:
        # --------------
        # if ref_com

        lok = [rr[0] for rr in lrcom]
        ref_com = ds._generic_check._check_var(
            ref_com, 'ref_com',
            types=str,
            allowed=lok,
        )

        # -----------
        # safety check vs assumptions

        if isinstance(x0, str) and ref_com is coll.ddata[x0]['ref']:
            msg = (
                "For submesh=True, x0 '{x0}' should have no common ref!"
                f"\n\t- detected: {ref_com}"
            )
            raise NotImplementedError(msg)

    return ref_com


def _submesh_addtemp(
    coll=None,
    kd0=None,
    # interp resut
    dout_temp=None,
    # coordinates
    x0=None,
):

    # ----------------
    # add ref and data

    if isinstance(x0, str):

       assert not any([rr is None for rr in dout_temp[kd0[0]]['ref']])
       dr_add, lr_add = None, None

    else:

        # dref
        ii, dr_add, lref = 0, {}, []
        for jj, rr in enumerate(dout_temp[kd0[0]]['ref']):

            # ref name
            if rr is None:
                rr = f"r{len(coll.dref) + ii:03.0f}"
                lref.append(rr)
                ii += 1
            else:
                lref.append(None)

            # dr_add
            if rr not in coll.dref.keys():
                dr_add[rr] = {
                    'key': rr,
                    'size': dout_temp[kd0[0]]['data'].shape[jj],
                }

        # combine
        for kk in dout_temp.keys():
            dout_temp[kk]['ref'] = tuple([
                rr if rr is not None else dout_temp[kd0[0]]['ref'][jj]
                for jj, rr in enumerate(lref)
            ])

        lr_add = [vv['key'] for vv in dr_add.values()]

    # --------------
    # populate dd_add

    dd_add = dout_temp
    for ii, kk in enumerate(dd_add.keys()):
        dd_add[kk]['key'] = f'dddd{len(coll.ddata) + ii:03.0f}'
    ld_add = [vv['key'] for vv in dd_add.values()]

    # --------
    # add

    # ref
    if dr_add is not None:
        for vv in dr_add.values():
            coll.add_ref(**vv)

    # data
    for vv in dd_add.values():
        coll.add_data(**vv)

    return lr_add, ld_add


# ################################################################
# ################################################################
#                       Prepare bsplines
# ################################################################


def _prepare_bsplines(
    coll=None,
    ref_key=None,
):

    # ----------------
    # keys and types

    wbs = coll._which_bsplines
    keybs = ref_key
    keym = coll.dobj[wbs][keybs]['mesh']
    ref_key = coll.dobj[wbs][keybs]['apex']
    mtype = coll.dobj[coll._which_mesh][keym]['type']

    # -------------
    # azone

    # azone = ds._generic_check._check_var(
        # azone, 'azone',
        # types=bool,
        # default=True,
    # )

    # -----------
    # indbs

    return keybs, keym, mtype, ref_key


# #############################################################################
# #############################################################################
#                          core interpolation
# #############################################################################


def _interp(
    coll=None,
    keybs=None,
    keys=None,
    x0=None,
    x1=None,
    val_out=None,
    deriv=None,
    indbs_tf=None,
    crop=None,
    cropbs=None,
    details=None,
    dout=None,
    mtype=None,
    ddata=None,
    daxis=None,
    sli_c=None,
    sli_x=None,
    sli_v=None,
    dsh_other=None,
    dref_com=None,
    lr_add=None,
    ld_add=None,
    nan0=None,
):

    # ----------
    # prepare

    # indokx
    indokx0 = np.isfinite(x0)
    if x1 is not None:
        indokx0 &= np.isfinite(x1)

    # loop on keys
    derr = {}
    clas = coll.dobj[coll._which_bsplines][keybs]['class']
    for ii, k0 in enumerate(keys):

        try:

            if details is True:
                dout[k0]['data'] = clas.ev_details(
                    x0=x0,
                    x1=x1,
                    # options
                    val_out=val_out,
                    deriv=deriv,
                    # indices
                    indbs_tf=indbs_tf,
                    # rect-specific
                    crop=crop,
                    cropbs=cropbs,
                )

            else:

                # --------------------
                # prepare slicing func

                if mtype == 'tri':
                    shape_c = ddata[k0].shape

                    (
                        shape_v, axis_v, ind_c, ind_v, shape_o,
                    ) = _utils_bsplines._get_shapes_ind(
                        axis=daxis[k0],
                        shape_c=shape_c,
                        shape_x=x0.shape,
                    )

                    sli_c = _utils_bsplines._get_slice_cx(
                        axis=daxis[k0],
                        shape=shape_c,
                        ind_cv=ind_c,
                        reverse=True,
                    )

                    sli_v = _utils_bsplines._get_slice_cx(
                        axis=axis_v,
                        shape=shape_v,
                        ind_cv=ind_v,
                        reverse=True,
                    )
                else:
                    axis_v = None

                sli_o = _utils_bsplines._get_slice_out(
                    axis=daxis[k0],
                    shape_c=ddata[k0].shape,
                )

                # --------------------
                # actual interpolation

                dout[k0]['data'][...] = clas(
                    x0=x0,
                    x1=x1,
                    # coefs
                    coefs=ddata[k0],
                    axis=daxis[k0],
                    # options
                    val_out=val_out,
                    deriv=deriv,
                    # rect-specific
                    crop=crop,
                    cropbs=cropbs,
                    # slicing
                    sli_c=sli_c,
                    sli_x=sli_x,
                    sli_v=sli_v,
                    sli_o=sli_o,
                    indokx0=indokx0,
                    shape_o=dsh_other[k0],
                    shape_v=dout[k0]['data'].shape,
                    dref_com=dref_com[k0],
                    axis_v=axis_v,
                )

                # nan0
                if nan0 is True:
                      dout[k0]['data'][dout[k0]['data'] == 0.] = np.nan

        except Exception as err:
            raise err
            derr[k0] = str(err)

    # ----------------------------
    # raise warning if any failure

    if len(derr) > 0:
        lstr = [f"\t- '{k0}': {v0}" for k0, v0 in derr.items()]
        msg = (
            "The following keys could not be interpolated:\n"
            + "\n".join(lstr)
        )
        warnings.warn(msg)

    # ----------------------------
    # clean-up temporary storage

    if lr_add is not None:
        for rr in lr_add:
            coll.remove_ref(rr, propagate=False)

    if ld_add is not None:
        for dd in ld_add:
            if dd in coll.ddata.keys():
                coll.remove_data(dd, propagate=False)

    # ----------
    # adjust ref

    if lr_add is not None:
        for k0, v0 in dout.items():
            dout[k0]['ref'] = tuple([
                None if rr in lr_add else rr
                for rr in v0['ref']
            ])

    if details:
        wbs = coll._which_bsplines
        for k0, v0 in dout.items():
            dout[k0]['ref'] = tuple(np.r_[
                v0['ref'], coll.dobj[wbs][keybs]['ref-bs'],
            ])

    return


# #############################################################################
# #############################################################################
#                           Mesh2DRect - interp
# #############################################################################


# def _interp2d_check_RZ(
    # R=None,
    # Z=None,
    # grid=None,
# ):
    # # R and Z provided
    # if not isinstance(R, np.ndarray):
        # try:
            # R = np.atleast_1d(R).astype(float)
        # except Exception as err:
            # msg = "R must be convertible to np.arrays of floats"
            # raise Exception(msg)
    # if not isinstance(Z, np.ndarray):
        # try:
            # Z = np.atleast_1d(Z).astype(float)
        # except Exception as err:
            # msg = "Z must be convertible to np.arrays of floats"
            # raise Exception(msg)

    # # grid
    # grid = ds._generic_check._check_var(
        # grid, 'grid',
        # default=R.shape != Z.shape,
        # types=bool,
    # )

    # if grid is True and (R.ndim > 1 or Z.ndim > 1):
        # msg = "If grid=True, R and Z must be 1d!"
        # raise Exception(msg)
    # elif grid is False and R.shape != Z.shape:
        # msg = "If grid=False, R and Z must have the same shape!"
        # raise Exception(msg)

    # if grid is True:
        # R = np.tile(R, Z.size)
        # Z = np.repeat(Z, R.size)
    # return R, Z, grid


# def _interp2d_check(
    # # ressources
    # coll=None,
    # # interpolation base, 1d or 2d
    # key=None,
    # # external coefs (optional)
    # coefs=None,
    # # interpolation points
    # R=None,
    # Z=None,
    # radius=None,
    # angle=None,
    # grid=None,
    # radius_vs_time=None,
    # azone=None,
    # # time: t or indt
    # t=None,
    # indt=None,
    # indt_strict=None,
    # # bsplines
    # indbs=None,
    # # parameters
    # details=None,
    # reshape=None,
    # res=None,
    # crop=None,
    # nan0=None,
    # val_out=None,
    # imshow=None,
    # return_params=None,
    # store=None,
    # inplace=None,
# ):

    # # -------------
    # # keys

    # dk = {
        # k0: v0['bsplines']
        # for k0, v0 in coll.ddata.items()
        # if v0.get('bsplines') not in ['', None]
        # and 'crop' not in k0
    # }
    # dk.update({kk: kk for kk in coll.dobj['bsplines'].keys()})
    # if key is None and len(dk) == 1:
        # key = list(dk.keys())[0]
    # if key not in dk.keys():
        # msg = (
            # "Arg key must the key to a data referenced on a bsplines set\n"
            # f"\t- available: {list(dk.keys())}\n"
            # f"\t- provided: {key}\n"
        # )
        # raise Exception(msg)

    # keybs = dk[key]
    # keym = coll.dobj['bsplines'][keybs]['mesh']
    # mtype = coll.dobj[coll._which_mesh][keym]['type']

    # # -------------
    # # details

    # details = ds._generic_check._check_var(
        # details, 'details',
        # types=bool,
        # default=False,
    # )

    # # -------------
    # # crop

    # crop = ds._generic_check._check_var(
        # crop, 'crop',
        # types=bool,
        # default=mtype == 'rect',
        # allowed=[False, True] if mtype == 'rect' else [False],
    # )

    # # -------------
    # # nan0

    # nan0 = ds._generic_check._check_var(
        # nan0, 'nan0',
        # types=bool,
        # default=True,
    # )

    # # -------------
    # # val_out

    # val_out = ds._generic_check._check_var(
        # val_out, 'val_out',
        # default=np.nan,
        # allowed=[False, np.nan, 0.]
    # )

    # # -----------
    # # time

    # # refbs and hastime
    # refbs = coll.dobj['bsplines'][keybs]['ref']

    # if mtype == 'polar':
        # radius2d = coll.dobj[coll._which_mesh][keym]['radius2d']
        # # take out key if bsplines
        # lk = [kk for kk in [key, radius2d] if kk != keybs]

        # #
        # hastime, reft, keyt, t, dind = coll.get_time_common(
            # keys=lk,
            # t=t,
            # indt=indt,
            # ind_strict=indt_strict,
        # )

        # rad2d_hastime = radius2d in dind.keys()
        # if key == keybs:
            # kind = radius2d
        # else:
            # kind = key
            # hastime = key in dind.keys()

        # if hastime:

            # indt = dind[kind].get('ind')

            # # Special case: all times match
            # if indt is not None:
                # rtk = coll.get_time(key)[2]
                # if indt.size == coll.dref[rtk]['size']:
                    # if np.allclose(indt, np.arange(0, coll.dref[rtk]['size'])):
                        # indt = None

            # if indt is not None:
                # indtu = np.unique(indt)
                # indtr = np.array([indt == iu for iu in indtu])
            # else:
                # indtu, indtr = None, None
        # else:
            # indt, indtu, indtr = None, None, None

    # elif key != keybs:
        # # hastime, t, indit
        # hastime, hasvect, reft, keyt, t, dind = coll.get_time(
            # key=key,
            # t=t,
            # indt=indt,
            # ind_strict=indt_strict,
        # )
        # if dind is None:
            # indt, indtu, indtr = None, None, None
        # else:
            # indt, indtu, indtr = dind['ind'], dind['indu'], dind['indr']
    # else:
        # hastime, hasvect = False, False
        # reft, keyt, indt, indtu, indtr = None, None, None, None, None

    # # -----------
    # # indbs

    # if details is True:
        # if mtype == 'rect':
            # returnas = 'tuple-flat'
        # elif mtype == 'tri':
            # returnas = int
        # elif mtype == 'polar':
            # if len(coll.dobj['bsplines'][keybs]['shape']) == 2:
                # returnas = 'array-flat'
            # else:
                # returnas = int

        # # compute validated indbs array with appropriate form
        # indbs_tf = coll.select_ind(
            # key=keybs,
            # returnas=returnas,
            # ind=indbs,
        # )
    # else:
        # indbs_tf = None

    # # -----
    # # store

    # store = ds._generic_check._check_var(
        # store, 'store',
        # types=bool,
        # default=False,
        # allowed=[False] if details else [True, False],
    # )

    # # -----------
    # # coordinates

    # # (R, Z) vs (radius, angle)
    # lc = [
        # R is None and Z is None and radius is None,
        # R is not None and Z is not None and store is False,
        # (R is None and Z is None)
        # and (radius is not None and mtype == 'polar') and store is False
    # ]

    # if not any(lc):
        # msg = (
            # "Please provide either:\n"
            # "\t- R and Z: for any mesh type\n"
            # "\t- radius (and angle): for polar mesh\n"
            # "\t- None: for rect / tri mesh\n"
        # )
        # raise Exception(msg)

    # # R, Z
    # if lc[0]:

        # if store is True:
            # if imshow:
                # msg = "Storing only works if imshow = False"
                # raise Exception(msg)

        # # no spec => sample mesh
        # R, Z = coll.get_sample_mesh(
            # key=keym,
            # res=res,
            # mode='abs',
            # grid=True,
            # R=R,
            # Z=Z,
            # imshow=imshow,
        # )
        # lc[1] = True

    # if lc[1]:

        # # check R, Z
        # R, Z, grid = _interp2d_check_RZ(R=R, Z=Z, grid=grid)

        # # special case if polar mesh => (radius, angle) from (R, Z)
        # if mtype == 'polar':

            # rad2d_indt = dind[radius2d].get('ind') if rad2d_hastime else None

            # # compute radius2d at relevant times
            # radius, _, _ = coll.interpolate_profile2d(
                # # coordinates
                # R=R,
                # Z=Z,
                # grid=False,
                # # quantities
                # key=radius2d,
                # details=False,
                # # time
                # indt=rad2d_indt,
            # )

            # # compute angle2d at relevant times
            # angle2d = coll.dobj[coll._which_mesh][keym]['angle2d']
            # if angle2d is not None:
                # angle, _, _ = coll.interpolate_profile2d(
                    # # coordinates
                    # R=R,
                    # Z=Z,
                    # grid=False,
                    # # quantities
                    # key=angle2d,
                    # details=False,
                    # # time
                    # indt=rad2d_indt,
                # )

            # # simplify if not time-dependent
            # if radius2d not in dind:
                # assert radius.ndim == R.ndim
                # if angle2d is not None:
                    # assert angle.ndim == R.ndim

            # # check consistency
            # if rad2d_hastime != (radius.ndim == R.ndim + 1):
                # msg = f"Inconsistency! {radius.shape}"
                # raise Exception(msg)

            # radius_vs_time = rad2d_hastime

        # else:
            # radius_vs_time = False

    # else:
        # radius_vs_time = ds._generic_check._check_var(
            # radius_vs_time, 'radius_vs_time',
            # types=bool,
            # default=False,
        # )

    # # -------------
    # # radius, angle

    # if mtype == 'polar':

        # # check same shape
        # if not isinstance(radius, np.ndarray):
            # radius = np.atleast_1d(radius)

        # # angle vs angle2d
        # angle2d = coll.dobj[coll._which_mesh][keym]['angle2d']
        # if angle2d is not None and angle is None:
            # msg = (
                # f"Arg angle must be provided for bsplines {keybs}"
            # )
            # raise Exception(msg)

        # # angle vs radius
        # if angle is not None:
            # if not isinstance(angle, np.ndarray):
                # angle = np.atleast_1d(angle)

            # if radius.shape != angle.shape:
                # msg = (
                    # "Args radius and angle must be np.ndarrays of same shape!\n"
                    # f"\t- radius.shape: {radius.shape}\n"
                    # f"\t- angle.shape: {angle.shape}\n"
                # )
                # raise Exception(msg)

    # # -------------
    # # coefs

    # shapebs = coll.dobj['bsplines'][keybs]['shape']
    # # 3 possible coefs shapes:
    # #   - None (if details = True or key provided)
    # #   - scalar or shapebs if details = False and key = keybs

    # if details is True:
        # coefs = None
        # axis = None
        # assert key == keybs, (key, keybs)

    # else:
        # if coefs is None:
            # if key == keybs:
                # if mtype == 'polar' and rad2d_hastime:
                    # if rad2d_indt is None:
                        # r2dnt = coll.dref[dind[radius2d]['ref']]['size']
                    # else:
                        # r2dnt = rad2d_indt.size
                    # coefs = np.ones(tuple(np.r_[r2dnt, shapebs]), dtype=float)
                    # axis = [ii + 1 for ii in range(len(shapebs))]
                # else:
                    # coefs = np.ones(shapebs, dtype=float)
                    # axis = [ii for ii in range(len(shapebs))]
            # else:
                # coefs = coll.ddata[key]['data']
                # axis = [
                    # ii for ii in range(len(coll.ddata[key]['ref']))
                    # if coll.ddata[key]['ref'][ii] in refbs
                # ]

        # elif key != keybs:
            # msg = f"Arg coefs can only be provided if key = keybs!\n\t- key: {key}"
            # raise Exception(msg)

        # elif np.isscalar(coefs):
            # coefs = np.full(shapebs, coefs)
            # axis = [ii for ii in range(len(shapebs))]

        # # consistency
        # nshbs = len(shapebs)
        # c0 = (
            # coefs.shape[-nshbs:] == shapebs
            # and (
                # (hastime and coefs.ndim == nshbs + 1)           # pre-shaped
                # or (not hastime and coefs.ndim == nshbs + 1)    # pre-shaped
                # or (not hastime and coefs.ndim == nshbs)        # [None, ...]
            # )
            # and len(axis) == len(refbs)
        # )
        # if not c0:
            # msg = (
                # f"Inconsistency of '{key}' shape:\n"
                # f"\t- shape: {coefs.shape}\n"
                # f"\t- shapebs: {shapebs}\n"
                # f"\t- hastime: {hastime}\n"
                # f"\t- radius_vs_time: {radius_vs_time}\n"
                # f"\t- refbs: {refbs}\n"
                # f"\t- axis: {axis}"
            # )
            # raise Exception(msg)

        # # Make sure coefs is time dependent
        # if hastime:
            # if indt is not None and (mtype == 'polar' or indtu is None):
                # if coefs.shape[0] != indt.size:
                    # # in case coefs is already provided with indt
                    # coefs = coefs[indt, ...]

            # if radius_vs_time and coefs.shape[0] != radius.shape[0]:
                # msg = (
                    # "Inconstistent coefs vs radius!\n"
                    # f"\t- coefs.shape = {coefs.shape}\n"
                    # f"\t- radius.shape = {radius.shape}\n"
                # )
                # raise Exception(msg)
        # else:
            # if radius_vs_time is True:
                # sh = tuple([radius.shape[0]] + [1]*len(shapebs))
                # coefs = np.tile(coefs, sh)
                # axis = [aa + 1 for aa in axis]
            # elif coefs.ndim == nshbs:
                # coefs = coefs[None, ...]
                # axis = [aa + 1 for aa in axis]

    # # -------------
    # # azone

    # azone = ds._generic_check._check_var(
        # azone, 'azone',
        # types=bool,
        # default=True,
    # )

    # # -------------
    # # return_params

    # return_params = ds._generic_check._check_var(
        # return_params, 'return_params',
        # types=bool,
        # default=False,
    # )

    # # -------
    # # inplace

    # inplace = ds._generic_check._check_var(
        # inplace, 'inplace',
        # types=bool,
        # default=store,
    # )

    # return (
        # key, keybs,
        # R, Z,
        # radius, angle,
        # coefs,
        # axis,
        # hastime,
        # reft, keyt,
        # shapebs,
        # radius_vs_time,
        # azone,
        # indbs, indbs_tf,
        # t, indt, indtu, indtr,
        # details, crop,
        # nan0, val_out,
        # return_params,
        # store, inplace,
    # )


# def interp2d(
    # # ressources
    # coll=None,
    # # interpolation base, 1d or 2d
    # key=None,
    # # external coefs (instead of key, optional)
    # coefs=None,
    # # interpolation points
    # R=None,
    # Z=None,
    # radius=None,
    # angle=None,
    # grid=None,
    # radius_vs_time=None,        # if radius is provided, in case radius vs time
    # azone=None,                 # if angle2d is interpolated, exclusion zone
    # # time: t or indt
    # t=None,
    # indt=None,
    # indt_strict=None,
    # # bsplines
    # indbs=None,
    # # parameters
    # details=None,
    # reshape=None,
    # res=None,
    # crop=None,
    # nan0=None,
    # val_out=None,
    # imshow=None,
    # return_params=None,
    # store=None,
    # inplace=None,
# ):

    # # ---------------
    # # check inputs

    # (
        # key, keybs,
        # R, Z,
        # radius, angle,
        # coefs,
        # axis,
        # hastime,
        # reft, keyt,
        # shapebs,
        # radius_vs_time,
        # azone,
        # indbs, indbs_tf,
        # t, indt, indtu, indtr,
        # details, crop,
        # nan0, val_out,
        # return_params,
        # store, inplace,
    # ) = _interp2d_check(
        # # ressources
        # coll=coll,
        # # interpolation base, 1d or 2d
        # key=key,
        # # external coefs (optional)
        # coefs=coefs,
        # # interpolation points
        # R=R,
        # Z=Z,
        # radius=radius,
        # angle=angle,
        # grid=grid,
        # radius_vs_time=radius_vs_time,
        # azone=azone,
        # # time: t or indt
        # t=t,
        # indt=indt,
        # indt_strict=indt_strict,
        # # bsplines
        # indbs=indbs,
        # # parameters
        # details=details,
        # res=res,
        # crop=crop,
        # nan0=nan0,
        # val_out=val_out,
        # imshow=imshow,
        # return_params=return_params,
        # store=store,
        # inplace=inplace,
    # )
    # keym = coll.dobj['bsplines'][keybs]['mesh']
    # meshtype = coll.dobj['mesh'][keym]['type']

    # # ----------------------
    # # which function to call

    # if details is True:
        # fname = 'func_details'
    # elif details is False:
        # fname = 'func_sum'
    # else:
        # raise Exception("Unknown details!")

    # # ---------------
    # # cropping ?

    # cropbs = coll.dobj['bsplines'][keybs]['crop']
    # if cropbs not in [None, False]:
        # cropbs = coll.ddata[cropbs]['data']
        # nbs = cropbs.sum()
    # else:
        # nbs = np.prod(coll.dobj['bsplines'][keybs]['shape'])

    # if indbs_tf is not None:
        # if isinstance(indbs_tf, tuple):
            # nbs = indbs_tf[0].size
        # else:
            # nbs = indbs_tf.size

    # # -----------
    # # Interpolate

    # if meshtype in ['rect', 'tri']:

        # # manage time
        # if indtu is not None:
            # val0 = np.full(tuple(np.r_[indt.size, R.shape]), np.nan)
            # coefs = coefs[indtu, ...]

        # val = coll.dobj['bsplines'][keybs][fname](
            # R=R,
            # Z=Z,
            # coefs=coefs,
            # axis=axis,
            # crop=crop,
            # cropbs=cropbs,
            # indbs_tf=indbs_tf,
            # val_out=val_out,
        # )

        # # manage time
        # if indtu is not None:
            # for ii, iu in enumerate(indtu):
                # val0[indtr[ii], ...] = val[ii, ...]
            # val = val0

        # shape_pts = R.shape

    # elif meshtype == 'polar':

        # val = coll.dobj['bsplines'][keybs][fname](
            # radius=radius,
            # angle=angle,
            # coefs=coefs,
            # axis=axis,
            # indbs_tf=indbs_tf,
            # radius_vs_time=radius_vs_time,
            # val_out=val_out,
        # )

        # shape_pts = radius.shape

    # # ---------------
    # # post-treatment for angle2d only (discontinuity)

    # # if azone is True:
        # # lkm0 = [
            # # k0 for k0, v0 in coll.dobj[coll._which_mesh].items()
            # # if key == v0.get('angle2d')
        # # ]
        # # if len(lkm0) > 0:
            # # ind = angle2d_inzone(
                # # coll=coll,
                # # keym0=lkm0[0],
                # # keya2d=key,
                # # R=R,
                # # Z=Z,
                # # t=t,
                # # indt=indt,
            # # )
            # # assert val.shape == ind.shape
            # # val[ind] = np.pi

    # # ---------------
    # # post-treatment

    # if nan0 is True:
        # val[val == 0] = np.nan

    # if not hastime and not radius_vs_time:
        # c0 = (
            # (
                # details is False
                # and val.shape == tuple(np.r_[1, shape_pts])
            # )
            # or (
                # details is True
                # and val.shape == tuple(np.r_[shape_pts, nbs])
            # )
        # )
        # if not c0:
            # import pdb; pdb.set_trace()     # DB
            # pass
        # if details is False:
            # val = val[0, ...]
            # reft = None

    # # ------
    # # store

    # # ref
    # ct = (
        # (hastime or radius_vs_time)
        # and (
            # reft in [None, False]
            # or (
                # reft not in [None, False]
                # and indt is not None
                # and not (
                    # indt.size == coll.dref[reft]['size']
                    # or np.allclose(indt, np.arange(0, coll.dref[reft]['size']))
                # )
            # )
        # )
    # )
    # if ct:
        # reft = f'{key}-nt'

    # # store
    # if store is True:
        # Ru = np.unique(R)
        # Zu = np.unique(Z)
        # nR, nZ = Ru.size, Zu.size

        # knR, knZ = f'{key}-nR', f'{key}-nZ'
        # kR, kZ = f'{key}-R', f'{key}-Z'

        # if inplace is True:
            # coll2 = coll
        # else:
            # coll2 = ds.DataStock()

        # # add ref nR, nZ
        # coll2.add_ref(key=knR, size=nR)
        # coll2.add_ref(key=knZ, size=nZ)

        # # add data Ru, Zu
        # coll2.add_data(
            # key=kR,
            # data=Ru,
            # ref=knR,
            # dim='distance',
            # name='R',
            # units='m',
        # )
        # coll2.add_data(
            # key=kZ,
            # data=Zu,
            # ref=knZ,
            # dim='distance',
            # name='Z',
            # units='m',
        # )

        # # ref
        # if hastime or radius_vs_time:
            # if ct:
                # coll2.add_ref(key=reft, size=t.size)
                # coll2.add_data(
                    # key=f'{key}-t',
                    # data=t,
                    # ref=reft,
                    # dim='time',
                    # units='s',
                # )
            # else:
                # if reft not in coll2.dref.keys():
                    # coll2.add_ref(key=reft, size=coll.dref[reft]['size'])
                # if keyt is not None and keyt not in coll2.ddata.keys():
                    # coll2.add_data(
                        # key=keyt,
                        # data=coll.ddata[keyt]['data'],
                        # dim='time',
                    # )

            # ref = (reft, knR, knZ)
        # else:
            # ref = (knR, knZ)

        # coll2.add_data(
            # data=val,
            # key=f'{key}-map',
            # ref=ref,
            # dim=coll.ddata[key]['dim'],
            # quant=coll.ddata[key]['quant'],
            # name=coll.ddata[key]['name'],
            # units=coll.ddata[key]['units'],
        # )

    # else:

        # ref = []
        # c0 = (
            # reft not in [None, False]
            # and (hastime or radius_vs_time)
            # and not (
                # meshtype == 'polar'
                # and R is None
                # and coefs is None
                # and radius_vs_time is False
            # )
        # )
        # if c0:
            # ref.append(reft)

        # if meshtype in ['rect', 'tri']:
            # for ii in range(R.ndim):
                # ref.append(None)
            # if grid is True:
                # for ii in range(Z.ndim):
                    # ref.append(None)
        # else:
            # for ii in range(radius.ndim):
                # if radius_vs_time and ii == 0:
                    # continue
                # ref.append(None)
            # if grid is True and angle is not None:
                # for ii in range(angle.ndim):
                    # ref.append(None)

        # if details is True:
            # refbs = coll.dobj['bsplines'][keybs]['ref-bs'][0]
            # if crop is True:
                # refbs = f"{refbs}-crop"
            # ref.append(refbs)
        # ref = tuple(ref)

        # if ref[0] == 'emiss-nt':
            # import pdb; pdb.set_trace()     # DB

        # if len(ref) != val.ndim:
            # msg = (
                # "Mismatching ref vs val.shape:\n"
                # f"\t- key = {key}\n"
                # f"\t- keybs = {keybs}\n"
                # f"\t- val.shape = {val.shape}\n"
                # f"\t- ref = {ref}\n"
                # f"\t- reft = {reft}\n"
                # f"\t- hastime = {hastime}\n"
                # f"\t- radius_vs_time = {radius_vs_time}\n"
                # f"\t- details = {details}\n"
                # f"\t- indbs_tf = {indbs_tf}\n"
                # f"\t- key = {key}\n"
                # f"\t- meshtype = {meshtype}\n"
                # f"\t- grid = {grid}\n"
            # )
            # if coefs is not None:
                # msg += f"\t- coefs.shape = {coefs.shape}\n"
            # if R is not None:
                # msg += (
                    # f"\t- R.shape = {R.shape}\n"
                    # f"\t- Z.shape = {Z.shape}\n"
                # )
            # if meshtype == 'polar':
                # msg += f"\t- radius.shape = {radius.shape}\n"
                # if angle is not None:
                    # msg += f"\t- angle.shape = {angle.shape}\n"
            # raise Exception(msg)

    # # ------
    # # return

    # if store and inplace is False:
        # return coll2
    # else:
        # if return_params is True:
            # return val, t, ref, dparams
        # else:
            # return val, t, ref


# #############################################################################
# #############################################################################
#                   2d points to 1d quantity interpolation
# #############################################################################


# def interp2dto1d(
    # coll=None,
    # key=None,
    # R=None,
    # Z=None,
    # indbs=None,
    # indt=None,
    # grid=None,
    # details=None,
    # reshape=None,
    # res=None,
    # crop=None,
    # nan0=None,
    # imshow=None,
    # return_params=None,
# ):

    # # ---------------
    # # check inputs

    # # TBD
    # pass

    # # ---------------
    # # post-treatment

    # if nan0 is True:
        # val[val == 0] = np.nan

    # # ------
    # # return

    # if return_params is True:
        # return val, dparams
    # else:
        # return val


# #############################################################################
# #############################################################################
#                           Mesh2DRect - operators
# #############################################################################


def get_bsplines_operator(
    coll,
    key=None,
    operator=None,
    geometry=None,
    crop=None,
    store=None,
    returnas=None,
    # specific to deg = 0
    centered=None,
    # to return gradR, gradZ, for D1N2 deg 0, for tomotok
    returnas_element=None,
):

    # check inputs
    lk = list(coll.dobj.get('bsplines', {}).keys())
    key = ds._generic_check._check_var(
        key, 'key',
        types=str,
        allowed=lk,
    )

    store = ds._generic_check._check_var(
        store, 'store',
        default=True,
        types=bool,
    )

    returnas = ds._generic_check._check_var(
        returnas, 'returnas',
        default=store is False,
        types=bool,
    )

    crop = ds._generic_check._check_var(
        crop, 'crop',
        default=True,
        types=bool,
    )

    # cropbs
    cropbs = coll.dobj['bsplines'][key]['crop']
    keycropped = coll.dobj['bsplines'][key]['ref-bs'][0]
    if cropbs not in [None, False] and crop is True:
        cropbs_flat = coll.ddata[cropbs]['data'].ravel(order='F')
        if coll.dobj['bsplines'][key]['deg'] == 0:
            cropbs = coll.ddata[cropbs]['data']
        keycropped = f"{keycropped}-crop"
    else:
        cropbs = False
        cropbs_flat = False

    # compute and return
    (
        opmat, operator, geometry, dim,
    ) = coll.dobj['bsplines'][key]['class'].get_operator(
        operator=operator,
        geometry=geometry,
        cropbs_flat=cropbs_flat,
        # specific to deg=0
        cropbs=cropbs,
        centered=centered,
        # to return gradR, gradZ, for D1N2 deg 0, for tomotok
        returnas_element=returnas_element,
    )

    # cropping
    if operator == 'D1':
        ref = (keycropped, keycropped)
    elif operator == 'D0N1':
        ref = (keycropped,)
    elif 'N2' in operator:
        ref = (keycropped, keycropped)

    return opmat, operator, geometry, dim, ref, crop, store, returnas, key



# -*- coding: utf-8 -*-


# Common
import numpy as np
import datastock as ds
from matplotlib import path


# tofu
from . import _generic_mesh


# ##############################################################
# ##############################################################
#                       Main
# ##############################################################


def sample_mesh(
    coll=None,
    key=None,
    res=None,
    mode=None,
    x0=None,
    x1=None,
    grid=None,
    # options
    Dx0=None,
    Dx1=None,
    submesh=None,
    # output for 2d
    imshow=None,
    return_ind=None,
    in_mesh=None,
    # store
    store=None,
    kx0=None,
    kx1=None,
):

    # -------------
    # check inputs

    (
        key, nd, mtype,
        mode, res, submesh,
        grid, imshow, store,
    ) = _check(
        coll=coll,
        key=key,
        mode=mode,
        res=res,
        submesh=submesh,
        grid=grid,
        imshow=imshow,
        store=store,
        x0=x0,
        x1=x1,
    )

    # ------------
    # sample

    if nd == '1d':

        # check
        knots, Dx = _check_1d(
            coll=coll,
            key=key,
            Dx=Dx0,
            mode=mode,
        )

        # sample
        x0 = _sample_1d(
            res=res,
            mode=mode,
            knots=knots,
            Dx=Dx,
        )

        # prepare storing
        dref, ddata = _store_1d(
            coll=coll,
            key=key,
            x0=x0,
            # store
            store=store,
            kx0=kx0,
        )

    else:

        # check
        knots0, knots1, x0, x1, Dx0, Dx1 = _check_2d(
            coll=coll,
            key=key,
            x0=x0,
            x1=x1,
            Dx0=Dx0,
            Dx1=Dx1,
            mode=mode,
            mtype=mtype,
        )

        # sample
        x0, x1, ind = _sample_2d(
            coll=coll,
            key=key,
            res=res,
            mode=mode,
            knots0=knots0,
            knots1=knots1,
            x0=x0,
            x1=x1,
            Dx0=Dx0,
            Dx1=Dx1,
            # options
            grid=grid,
            imshow=imshow,
            in_mesh=in_mesh,
        )

        # prepare storing
        dref, ddata = _store_2d(
            coll=coll,
            key=key,
            x0=x0,
            x1=x1,
            ind=ind,
            # store
            store=store,
            kx0=kx0,
            kx1=kx1,
        )

    # -------------
    # format output

    # -------------
    # store

    if store is True:

        _store(coll=coll, dref=dref, ddata=ddata)

    # --------
    # return

    return ddata


# ##################################################################
# ##################################################################
#                       checks
# ##################################################################


def _check(
    coll=None,
    key=None,
    mode=None,
    res=None,
    submesh=None,
    grid=None,
    imshow=None,
    store=None,
    x0=None,
    x1=None,
):

    # ---------
    # mesh

    # submesh
    submesh = ds._generic_check._check_var(
        submesh, 'submesh',
        types=bool,
        default=False,
    )

    # key
    wm = coll._which_mesh
    key, _, cat = _generic_mesh._get_key_mesh_vs_bplines(
        coll=coll,
        key=key,
        which=wm,
    )

    nd = coll.dobj[cat][key]['nd']
    mtype = coll.dobj[cat][key]['type']

    if nd == '1d' and submesh is True:
        if coll.dobj[wm][key]['submesh'] is not None:
            key = coll.dobj[wm][key]['submesh']
            nd = coll.dobj[cat][key]['nd']
            mtype = coll.dobj[cat][key]['type']

    # ----------
    # resolution

    # mode
    mode = ds._generic_check._check_var(
        mode, 'mode',
        types=str,
        default='abs',
    )

    # res
    res = _get_res(
        coll=coll,
        key=key,
        res=res,
        nd=nd,
        mtype=mtype,
    )

    # ---------
    # parameters

    if nd == '1d':
        lok = None

    else:

        # grid
        grid = ds._generic_check._check_var(
            grid, 'grid',
            types=bool,
            default=False,
        )

        # imshow
        imshow = ds._generic_check._check_var(
            imshow, 'imshow',
            types=bool,
            default=False,
        )

        lok = [False]
        if grid is False and imshow is False:
            if x0 is None and x1 is None:
                lok.append(True)

    # --------
    # store

    store = ds._generic_check._check_var(
        store, 'store',
        types=bool,
        default=False,
        allowed=lok,
    )

    return (
        key, nd, mtype,
        mode, res, submesh,
        grid, imshow, store,
    )


def _check_1d(
    coll=None,
    key=None,
    Dx=None,
    mode=None,
):

    wm = coll._which_mesh

    # -------------
    # knots

    kknots = coll.dobj[wm][key]['knots'][0]
    knots = coll.ddata[kknots]['data']

    # custom DR or DZ for mode='abs' only
    Dx = _Dx(Dx)

    if Dx is not None and mode != 'abs':
        msg = "Custom Dx can only be provided with mode = 'abs'!"
        raise Exception(msg)

    return knots, Dx


def _check_2d(
    coll=None,
    key=None,
    x0=None,
    x1=None,
    Dx0=None,
    Dx1=None,
    mode=None,
    mtype=None,
):

    wm = coll._which_mesh

    # -----------
    # Parameters

    # R, Z
    if x0 is None and x1 is None:
        pass
    elif x0 is None and np.isscalar(x1):
        pass
    elif x1 is None and np.isscalar(x0):
        pass
    else:
        msg = (
            "For mesh discretisation, (x0, x1) can be either:\n"
            "\t- (None, None): will be created\n"
            "\t- (scalar, None): A vertical line will be created\n"
            "\t- (None, scalar): A horizontal line will be created\n"
        )
        raise Exception(msg)

    # -------------
    # x0, x1

    k0, k1 = coll.dobj[wm][key]['knots']
    if mtype == 'rect':
        knots0 = coll.ddata[k0]['data']
        knots1 = coll.ddata[k1]['data']

        # custom R xor Z for vertical / horizontal lines only
        if x0 is None and x1 is not None:
            x0 = knots0
        if x1 is None and x0 is not None:
            x1 = knots1

    else:
        knots0 = coll.ddata[k0]['data']
        knots1 = coll.ddata[k1]['data']

    # custom DR or DZ for mode='abs' only
    Dx0 = _DRZ(Dx=Dx0, size=None, Dx_name='Dx0')
    Dx1 = _DRZ(
        Dx=Dx1,
        size=len(Dx0) if Dx0 is not None else None,
        Dx_name='Dx1',
    )

    if (Dx0 is not None or Dx1 is not None) and mode != 'abs':
        msg = (
            "Custom Dx0 or Dx1 can only be provided with mode = 'abs'!\n"
            f"\t- mode: {mode}\n"
            f"\t- Dx0: {Dx0}\n"
            f"\t- Dx1: {Dx1}\n"
        )
        raise Exception(msg)

    return knots0, knots1, x0, x1, Dx0, Dx1


# #############
# check Dx
# #############


def _Dx(Dx=None):

    if Dx is None:
        return None

    c0 = (
        hasattr(Dx, '__iter__')
        and len(Dx) == 2
        and all([
            rr is None or (np.isscalar(rr) and not np.isnan(rr))
            for rr in Dx
        ])
    )
    if not c0:
        msg = 'Arg Dx must be an iterable of 2 scalars!'
        raise Exception(msg)

    for ii in range(2):
        if Dx[ii] is None:
            Dx[ii] = -np.inf if ii == 00 else np.inf

    return Dx


def _DRZ(Dx=None, size=None, Dx_name=None):

    if Dx is None:
        return None

    if None not in Dx and np.all(np.isfinite(Dx)) and np.size(Dx) > 2:
        Dx = np.atleast_1d(Dx).ravel()
        if size is not None and Dx.size != size:
            msg = (
                f"Arg {Dx_name} has wrong size!\n"
                f"\t- expected: {size}\n"
                f"\t- Provided: {Dx.size}"
            )
            raise Exception(msg)

    else:
        Dx = _Dx(Dx)

    return Dx


# ################################################################
# ################################################################
#                       utility
# ################################################################


def _get_res(
    coll=None,
    key=None,
    res=None,
    nd=None,
    mtype=None,
):

    wm = coll._which_mesh
    if nd == '1d':

        if res is None:
            kknots = coll.dobj[wm][key]['knots'][0]
            res = np.min(np.diff(coll.ddata[kknots]['data']))

        if not (np.isscalar(res) and res > 0.):
            msg = (
                f"Arg res must be a positive float!\n"
                "Provided: {res}"
            )
            raise Exception(msg)

    else:

        if res is None:
            if mtype == 'rect':
                kR, kZ = coll.dobj[wm][key]['knots']
                res = min(
                    np.min(np.diff(coll.ddata[kR]['data'])),
                    np.min(np.diff(coll.ddata[kZ]['data'])),
                )

            elif mtype == 'tri':
                res = 0.02

        # check len() = 2
        if np.isscalar(res):
            res = [res, res]

        c0 = (
            isinstance(res, list)
            and len(res) == 2
            and all([np.isscalar(rr) and rr > 0 for rr in res])
        )
        if not c0:
            msg = (
                f"Arg res must be a list of 2 positive floats!\n"
                "Provided: {res}"
            )
            raise Exception(msg)

    return res


# ##################################################################
# ##################################################################
#                       sample 1d
# ##################################################################


def _sample_1d(
    res=None,
    mode=None,
    knots=None,
    Dx=None,
):

    kmin, kmax = knots.min(), knots.max()

    if mode == 'abs':
        nx = int(np.ceil((kmax - kmin) / res))
        xx = np.linspace(kmin, kmax, nx)

    else:
        nx = int(np.ceil(1./res))
        kx = np.linspace(0, 1, nx, endpoint=False)[None, :]
        xx = np.concatenate((
            (knots[:-1, None] + kx*np.diff(knots)[:, None]).ravel(),
            knots[-1:],
        ))

    # Dx
    if Dx is not None:
        ind = (xx >= Dx[0]) & (xx < Dx[1])
        xx = xx[ind]

    return xx


# ##################################################################
# ##################################################################
#                       sample 2d
# ##################################################################


def _sample_2d(
    coll=None,
    key=None,
    res=None,
    mode=None,
    knots0=None,
    knots1=None,
    x0=None,
    x1=None,
    Dx0=None,
    Dx1=None,
    # options
    grid=None,
    imshow=None,
    in_mesh=None,
):

    # --------
    # compute

    min0, max0 = knots0.min(), knots0.max()
    min1, max1 = knots1.min(), knots1.max()

    if mode == 'abs':
        if x0 is None:
            n0 = int(np.ceil((max0 - min0) / res[0]))
            x0 = np.linspace(min0, max0, n0)
        if x1 is None:
            n1 = int(np.ceil((max1 - min1) / res[1]))
            x1 = np.linspace(min1, max1, n1)
    else:
        if x0 is None:
            n0 = int(np.ceil(1./res[0]))
            kx0 = np.linspace(0, 1, n0, endpoint=False)[None, :]
            x0 = np.concatenate((
                (knots0[:-1, None] + kx0*np.diff(knots0)[:, None]).ravel(),
                knots0[-1:],
            ))
        if x1 is None:
            n1 = int(np.ceil(1./res[1]))
            kx1 = np.linspace(0, 1, n1, endpoint=False)[None, :]
            x1 = np.concatenate((
                (knots1[:-1, None] + kx1*np.diff(knots1)[:, None]).ravel(),
                knots1[-1:],
            ))

    if np.isscalar(x0):
        x0 = np.full(x1.shape, x0)
    if np.isscalar(x1):
        x1 = np.full(x0.shape, x1)

    # -----------
    # prepare ind

    # x0, x1
    if grid or in_mesh or Dx0 is not None:
        x02 = np.repeat(x0[:, None], x1.size, axis=1)
        x12 = np.repeat(x1[None, :], x0.size, axis=0)

        if in_mesh or Dx0 is not None:
            x0f = x02.ravel()
            x1f = x12.ravel()
            sh = x02.shape

    # ind
    if Dx0 is not None or in_mesh is True:
        ind = np.ones((x0.size, x1.size), dtype=bool)
    else:
        ind = None

    # ---------
    # Dx0, Dx1

    if Dx0 is not None and len(Dx0) == 2:
        ind = ind & (
            (x02 >= Dx0[0]) & (x02 < Dx0[1])
            & (x12 >= Dx1[0]) & (x12 < Dx1[1])
        )

    # -----------
    # indices

    # mesh outline
    if in_mesh is True:
        dout = coll.get_mesh_outline(key=key)
        pa = path.Path(np.array([dout['x0']['data'], dout['x1']['data']]).T)
        ind = ind & pa.contains_points(np.array([x0f, x1f]).T).reshape(sh)

    # Dx0, Dx1
    if Dx0 is not None and len(Dx0) > 2:
        pa = path.Path(np.array([Dx0, Dx1]).T)
        ind = ind & pa.contains_points(np.array([x0f, x1f]).T).reshape(sh)

    # ------------
    # grid

    if grid is True:
        x0, x1 = x02, x12

    # ----------
    # imshow

    if imshow is True:
        if ind is not None:
            ind = ind.T
        if grid is True:
            x0 = x0.T
            x1 = x1.T

    return x0, x1, ind


# ##################################################################
# ##################################################################
#                       store
# ##################################################################


def _store_1d(
    coll=None,
    key=None,
    x0=None,
    # store
    store=None,
    kx0=None,
):

    # -----------
    # check key

    if store is True:
        lout = list(coll.ddata.keys())
    else:
        lout = []

    kx0 = ds._generic_check._check_var(
        kx0, 'kx0',
        types=str,
        default=f'{key}_x0',
        excluded=lout,
    )

    # -----------
    # format dict

    wm = coll._which_mesh
    kknots = coll.dobj[wm][key]['knots'][0]

    # dref
    kr = f'{kx0}_n'
    dref = {kr: {'size': x0.size}}

    # ddata
    ddata = {
        'x0': {
            'key': kx0,
            'data': x0,
            'dim': coll.ddata[kknots]['dim'],
            'quant': coll.ddata[kknots]['quant'],
            'name': coll.ddata[kknots]['name'],
            'units': coll.ddata[kknots]['units'],
        },
    }

    if store:
        ddata['x0']['ref'] = kr

    return dref, ddata


def _store_2d(
    coll=None,
    key=None,
    x0=None,
    x1=None,
    ind=None,
    # store
    store=None,
    kx0=None,
    kx1=None,
):


    # -----------
    # check key

    if store is True:
        lout = list(coll.ddata.keys())
    else:
        lout = []

    kx0 = ds._generic_check._check_var(
        kx0, 'kx0',
        types=str,
        default=f'{key}_x0_temp',
        excluded=lout,
    )

    kx1 = ds._generic_check._check_var(
        kx1, 'kx1',
        types=str,
        default=f'{key}_x1_temp',
        excluded=lout,
    )

    # -----------
    # format dict

    wm = coll._which_mesh
    kk0, kk1 = coll.dobj[wm][key]['knots']

    # dref
    k0r, k1r = f'{kx0}_n', f'{kx1}_n'
    dref = {
        k0r: {'size': x0.size},
        k1r: {'size': x1.size},
    }

    # ddata
    ddata = {
        'x0': {
            'key': kx0,
            'data': x0,
            'dim': coll.ddata[kk0]['dim'],
            'quant': coll.ddata[kk0]['quant'],
            'name': coll.ddata[kk0]['name'],
            'units': coll.ddata[kk0]['units'],
        },
        'x1': {
            'key': kx1,
            'data': x1,
            'dim': coll.ddata[kk1]['dim'],
            'quant': coll.ddata[kk1]['quant'],
            'name': coll.ddata[kk1]['name'],
            'units': coll.ddata[kk1]['units'],
        },
    }

    if ind is not None:
        ddata['ind'] = {
            'key': f'{key}_ind_temp',
            'data': ind,
        }

    if store is True:
        ddata['x0']['ref'] = k0r
        ddata['x1']['ref'] = k1r
        if ind is not None:
            ddata['ind']['ref'] = (k0r, k1r)

    return dref, ddata


def _store(coll=None, dref=None, ddata=None):

    # dref
    for k0, v0 in dref.items():
        coll.add_ref(key=k0, **v0)

    # ddata
    for k0, v0 in ddata.items():
        coll.add_data(**v0)



# -*- coding: utf-8 -*-


# Common
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import datastock as ds

# specific


# #############################################################################
# #############################################################################
#                           Main
# #############################################################################


def plot_as_profile2d(
    # ressources
    coll=None,
    # inputs
    key=None,
    # parameters
    dres=None,
    dunique_mesh_2d=None,
    # contours
    dlevels=None,
    ref_com=None,
    # details
    plot_details=None,
    # ref vectors
    dref_vectorZ=None,
    dref_vectorU=None,
    ref_vector_strategy=None,
    # interpolation
    val_out=None,
    nan0=None,
    # figure
    vmin=None,
    vmax=None,
    cmap=None,
    dax=None,
    dmargin=None,
    fs=None,
    dcolorbar=None,
    dleg=None,
    # interactivity
    dinc=None,
    connect=None,
):

    # --------------
    # check input

    if connect is None:
        connect = True

    (
        dkeys,
        dlevels,
        cmap, _,
        vmin, vmax,
        dcolorbar, dleg,
        connect,
    ) = _check(
        coll=coll,
        keys=key,
        dlevels=dlevels,
        # plotting
        vmin=vmin,
        vmax=vmax,
        cmap=cmap,
        # figure
        dcolorbar=dcolorbar,
        dleg=dleg,
        # interactivity
        connect=connect,
    )

    # --------------
    #  Prepare data

    (
        collax, dkeys,
        dlevels, lcol,
    ) = _prepare(
        coll=coll,
        dkeys=dkeys,
        dres=dres,
        dunique_mesh_2d=dunique_mesh_2d,
        # interpolation
        val_out=val_out,
        nan0=nan0,
        # levels
        dlevels=dlevels,
        ref_com=ref_com,
        # ref vectors
        dref_vectorZ=dref_vectorZ,
        dref_vectorU=dref_vectorU,
    )

    lkeys = ['key', 'keyX', 'keyY', 'keyZ', 'keyU']

    # --------------
    #  Prepare dax

    daxi = None
    if dax is not None:
        if not all([k0 in dax.keys() for k0 in dkeys.keys()]):
            dax = {k0: dax for k0 in dkeys.keys()}

    # nmax, color
    if len(dkeys) == 1:
        nmax = None
        color_dict = {k0: None for k0 in dkeys.keys()}
    else:
        nmax = 1
        color_dict = {k0: lcol[ii] for ii, k0 in enumerate(dkeys.keys())}

    # -----------------
    # loop on profile 2d

    for k0, v0 in dkeys.items():

        # -----------------
        # case with submesh

        if dax is not None:
            daxi = dax[k0]

        if v0['submesh'] is not None:
            collax, dgroup = _plot_submesh(
                coll=coll,
                collax=collax,
                key=k0,
                keym=v0['keym'],
                keybs=v0['keybs'],
                # ref vector
                dref_vector=dref_vectorZ,
                ref_vector_strategy=ref_vector_strategy,
                # details
                plot_details=plot_details,
                # plotting
                vmin=vmin,
                vmax=vmax,
                cmap=cmap,
                color_dict=color_dict[k0],
                # figure
                dax=daxi,
                dmargin=dmargin,
                fs=fs,
                dcolorbar=dcolorbar,
                dleg=dleg,
                interp=v0['interp'],
                dkeys={k1: v0[k1] for k1 in lkeys},
                lcol=lcol,
            )

        # -------------------
        # without submesh

        else:
            collax, dgroup = collax.plot_as_array(
                vmin=vmin,
                vmax=vmax,
                cmap=cmap,
                dax=daxi,
                dmargin=dmargin,
                fs=fs,
                dcolorbar=dcolorbar,
                dleg=dleg,
                interp=v0['interp'],
                color_dict=color_dict[k0],
                nmax=nmax,
                connect=False,
                inplace=True,
                **{k1: v0[k1] for k1 in lkeys},
            )

        # -----------
        # add levels

        if dlevels is not None:
            _add_levels_2d(
                collax=collax,
                dgroup=dgroup,
                dlevels=dlevels,
            )

    # -----------
    # connect

    if connect is True:
        collax.setup_interactivity(kinter='inter0', dgroup=dgroup, dinc=dinc)
        collax.disconnect_old()
        collax.connect()

        collax.show_commands()
        return collax
    else:
        return collax, dgroup


# #############################################################################
# #############################################################################
#                       Check
# #############################################################################


def _check(
    coll=None,
    keys=None,
    dlevels=None,
    # figure
    vmin=None,
    vmax=None,
    cmap=None,
    cmap_err=None,
    dcolorbar=None,
    dleg=None,
    # interactivity
    connect=None,
):

    # ----------
    # keys

    # key
    dk = coll.get_profiles2d()

    if isinstance(keys, str):
        keys = [keys]
    keys = ds._generic_check._check_var_iter(
        keys, 'keys',
        types=(list, tuple),
        types_iter=str,
        allowed=list(dk.keys()),
    )

    # check refs
    dkeys = {}
    wm = coll._which_mesh
    wbs = coll._which_bsplines
    for ii, k0 in enumerate(keys):

        keybs = dk[k0]
        keym = coll.dobj[wbs][keybs][wm]
        nd = coll.dobj[wm][keym]['nd']
        mtype = coll.dobj[wm][keym]['type']

        # submesh
        submesh = coll.dobj[wm][keym]['submesh']
        if submesh == '':
            submesh = None

        if submesh is None:
            subbs = keybs
            submtype = mtype
        else:
            subbs = coll.dobj[wm][keym]['subbs']
            # subkey = coll.dobj[wm][keym]['subkey']
            submtype = coll.dobj[wm][keym]['type']

        # ref_other
        ref_other = [
            rr for rr in coll.ddata[k0]['ref']
            if rr not in coll.dobj[wbs][keybs]['ref']
        ]
        if ii == 1:
            if ref_other != dkeys[keys[0]]['ref_other']:
                lstr = [f"\t- {kk}: {coll.ddata[kk]['ref']}" for kk in keys]
                msg = (
                    "Provided keys must be profile2d with identical ref!"
                    " (apart from the actual profile2d bsplines)\n"
                    f"\t- Provided:"
                    + "\n".join(lstr)
                )
                raise Exception(msg)

        dkeys[k0] = {
            'keybs': keybs,
            'keym': keym,
            'nd': nd,
            'mtype': mtype,
            'submesh': submesh,
            'subbs': subbs,
            'submtype': submtype,
            'ref_other': ref_other,
        }

    # ----------
    # dlevels

    if dlevels is not None:

        dp2d = coll.get_profiles2d()
        if isinstance(dlevels, (float, int)):
            dlevels = {k0: np.r_[dlevels] for k0 in keys}

        elif isinstance(dlevels, (np.ndarray, list, tuple)):
            dlevels = np.atleast_1d(dlevels).ravel()
            dlevels = {k0: dlevels for k0 in keys}

        c0 = (
            isinstance(dlevels, dict)
            and all([kk in dp2d.keys() for kk, vv in dlevels.items()])
        )
        if not c0:
            msg = (
                "Arg dlevels must be a dict with:\n"
                "\t- keys: valid keys of 2d profile data\n"
                "\t- values: iterable of level values\n"
                f"\nProvided: {dlevels}"
            )
            raise Exception(msg)

        for k0, v0 in dlevels.items():

            if isinstance(v0, (np.ndarray, list, tuple)):
                dlevels[k0] = {'levels': np.atleast_1d(v0).ravel()}
                v0 = dlevels[k0]

            if not isinstance(v0, dict) or 'levels' not in v0.keys():
                msg = f"dlevels['{k0}'] must have key 'levels'"
                raise Exception(msg)

            # check fields
            dlevels[k0]['levels'] = np.atleast_1d(v0['levels']).ravel()
            dlevels[k0]['color'] = v0.get('color', 'k')

    # ----------
    # cmap

    # cmap
    if cmap is None:
        cmap = 'viridis'

    # cmap_err
    if cmap_err is None:
        cmap_err = 'seismic'

    # ----------
    # vmin, vmax

    if vmin is None:
        vmin = min([0] + [np.nanmin(coll.ddata[kk]['data']) for kk in keys])
    if vmax is None:
        vmax = max([np.nanmax(coll.ddata[kk]['data']) for kk in keys])

    # ----------
    # figure

    # dcolorbar
    defdcolorbar = {
        # 'location': 'right',
        'fraction': 0.15,
        'orientation': 'vertical',
    }
    dcolorbar = ds._generic_check._check_var(
        dcolorbar, 'dcolorbar',
        default=defdcolorbar,
        types=dict,
    )

    # dleg
    defdleg = {
        'bbox_to_anchor': (1.1, 1.),
        'loc': 'upper left',
        'frameon': True,
    }
    dleg = ds._generic_check._check_var(
        dleg, 'dleg',
        default=defdleg,
        types=(bool, dict),
    )

    # ----------
    # interactivity

    connect = ds._generic_check._check_var(
        connect, 'connect',
        types=bool,
        default=True,
    )

    return (
        dkeys,
        dlevels,
        cmap, cmap_err,
        vmin, vmax,
        dcolorbar, dleg,
        connect,
    )


# #############################################################################
# #############################################################################
#                       Prepare
# #############################################################################


def _prepare(
    coll=None,
    dkeys=None,
    dres=None,
    dunique_mesh_2d=None,
    # levels
    dlevels=None,
    ref_com=None,
    # interpolation
    val_out=None,
    nan0=None,
    # ref vectors
    dref_vectorZ=None,
    dref_vectorU=None,
):

    # -----------
    # check

    if dref_vectorZ is None:
        dref_vectorZ = {}

    if dref_vectorU is None:
        dref_vectorU = {}

    # ---------------------------
    # get interpolated collection

    coll2, dbs = coll.interpolate_all_bsplines(
        keys=list(dkeys.keys()),
        dres=dres,
        dunique_mesh_2d=dunique_mesh_2d,
        submesh=True,
        # interpolation
        val_out=val_out,
        nan0=nan0,
    )
    lbs2d = [k0 for k0, v0 in dbs.items() if len(v0['ref']) == 2]

    # -----------------
    # get plotting mesh

    for k0, v0 in dkeys.items():

        # deg and interp
        wbs = coll._which_bsplines
        deg = coll.dobj[wbs][v0['subbs']]['deg']
        if deg == 0:
            interp = 'nearest'
        elif deg == 1:
            interp = 'bilinear'
        elif deg >= 2:
            interp = 'bicubic'

        dkeys[k0]['deg'] = deg
        dkeys[k0]['interp'] = interp

        # key X, Y, Z, U
        bs2d = [k1 for k1 in lbs2d if k1 in coll.ddata[k0][wbs]][0]
        rX, rY = dbs[bs2d]['ref']
        lr1d = [k0 for k0 in coll2.ddata[k0]['ref'] if k0 not in [rX, rY]]
        ndim = coll2.ddata[k0]['data'].ndim

        dkeys[k0].update({
            'key': k0,
            'keyX': coll2.get_ref_vector(ref=rX)[3],
            'keyY': coll2.get_ref_vector(ref=rY)[3],
            'keyZ': None,
            'keyU': None,
        })

        if ndim >= 3:
            dkeys[k0]['keyZ'] = coll2.get_ref_vector(
                ref=lr1d[0],
                **dref_vectorZ,
            )[3]
            # uniform = ds._plot_as_array._check_uniform_lin(
                # k0=keyZ, ddata=coll2.ddata,
            # )
            # if not uniform:
                # keyZ = None
            if ndim == 4:
                dkeys[k0]['keyU'] = coll2.get_ref_vector(
                    ref=lr1d[1],
                    **dref_vectorU,
                )[3]

    # -----------------
    # optional contours

    if dlevels is not None:

        for ii, (k0, v0) in enumerate(dlevels.items()):

            refi = coll.ddata[k0]['ref']

            # get contours
            dout, dref = coll.get_profile2d_contours(
                key=k0,
                levels=v0['levels'],
                ref_com=ref_com if ref_com in refi else None,
                res=dres if isinstance(dres, (int, float)) else None,
                store=False,
                return_dref=True,
                key_cont0="cont0",
                key_cont1="cont1",
            )

            # ref
            dlevels[k0]['dref'] = dref

            # axis
            ref = dout['cont0']['ref']
            axis = [
                ii for ii, rr in enumerate(ref)
                if rr not in coll.dref.keys()
            ]

            # refZ, refU
            refZ, refU = None, None
            if len(ref) == 3:
                refZ = [
                    rr for ii, rr in enumerate(ref)
                    if ii not in axis
                ][0]

            # populate
            for k1 in ['cont0', 'cont1']:
                dlevels[k0][k1] = dout[k1]
                dlevels[k0]['cont0']['key'] = f'{k0}_{k1}'

            dlevels[k0]['refZ'] = refZ
            dlevels[k0]['refU'] = refU
            dlevels[k0]['axis'] = axis

    # --------------------
    # optional for submesh

    lcol = ['k', 'b', 'g', 'r', 'm', 'c', 'y']

    return (
        coll2, dkeys,
        dlevels, lcol,
    )


# #############################################################################
# #############################################################################
#                   Plot fixed mesh with levels
# #############################################################################


def _add_levels_2d(
    collax=None,
    key=None,
    dgroup=None,
    dlevels=None,
):

    # ---------------------------------
    # add make contours as single lines

    ndim = len(dgroup)
    for ii, (k0, v0) in enumerate(dlevels.items()):

        for k1 in ['cont0', 'cont1']:

            v1 = v0[k1]
            sh = v1['data'].shape
            shnan = [1 if ii == v0['axis'][0] else ss for ii, ss in enumerate(sh)]

            dlevels[k0][k1]['data'] = np.append(
                v1['data'],
                np.full(tuple(shnan), np.nan),
                axis=v0['axis'][0],
            )

            sh = dlevels[k0][k1]['data'].shape
            newpts = sh[v0['axis'][0]]*sh[v0['axis'][1]]
            sh = tuple(np.r_[
                sh[:v0['axis'][0]],
                newpts,
                sh[v0['axis'][1]+1:]
            ].astype(int))

            newref = tuple(np.r_[
                v1['ref'][:v0['axis'][0]],
                [v0['dref']['npts']['key']],
                v1['ref'][v0['axis'][1]+1:],
            ])

            dlevels[k0][k1]['data'] = dlevels[k0][k1]['data'].swapaxes(
                v0['axis'][0],
                v0['axis'][1],
            ).reshape(sh)
            dlevels[k0][k1]['ref'] = newref

        dlevels[k0]['dref']['npts']['size'] = newpts

        if ii == 0:
            collax.add_ref(**dlevels[k0]['dref']['npts'])
            collax.add_ref(**dlevels[k0]['dref']['levels'])

        for k1 in ['cont0', 'cont1']:
            collax.add_data(**dlevels[k0][k1])

    # -----------
    # add contour

    kax = 'matrix'
    if collax.dax.get(kax) is not None:
        ax = collax.dax[kax]['handle']

        if ndim == 2:

            for ii, (k0, v0) in enumerate(dlevels.items()):
                ax.plot(
                    v0['cont0']['data'],
                    v0['cont1']['data'],
                    ls='-',
                    lw=1.,
                    c=dlevels[k0]['color'],
                )

        elif ndim == 3:

            for ii, (k0, v0) in enumerate(dlevels.items()):

                if v0['refZ'] is None:
                    ax.plot(
                        v0['cont0']['data'],
                        v0['cont1']['data'],
                        ls='-',
                        lw=1.,
                        c=dlevels[k0]['color'],
                    )

                else:
                    # slice
                    sli = [
                        slice(None) if ii == v0['axis'][0] else 0
                        for ii in range(ndim-1)
                    ]

                    # plot
                    l0, = ax.plot(
                        v0['cont0']['data'][sli],
                        v0['cont1']['data'][sli],
                        ls='-',
                        lw=1.,
                        c=dlevels[k0]['color'],
                    )

                    # store mobile
                    km = f'{k0}_contours'
                    collax.add_mobile(
                        key=km,
                        handle=l0,
                        # group_vis='Z',
                        # refs=[(refZ, ref_lvls), (refZ, ref_lvls)],
                        refs=[(v0['refZ'],), (v0['refZ'],)],
                        data=[v0['cont0']['key'], v0['cont1']['key']],
                        dtype=['xdata', 'ydata'],
                        axes=kax,
                        ind=0,
                    )

        else:
            raise NotImplementedError()

            # sli = [slice(None)]

            # l0, = ax.plot(
            # )

            # k0 = f'contour{ii}'
            # collax.add_mobile(
            # )

    # --------------------
    # add horizontal lines

    kax = 'radial'
    if collax.dax.get(kax) is not None and key in dlevels.keys():
        ax = collax.dax[kax]['handle']

        for ii, ll in enumerate(dlevels[key]['levels']):
            ax.axhline(ll, c='k', ls='--')


# #############################################################################
# #############################################################################
#                       Utilities
# #############################################################################


def _plot_bsplines_get_dx01(coll=None, km=None):
    # Get minimum distances

    wm = coll._which_mesh
    mtype = coll.dobj[wm][km]['type']
    if mtype == 'rect':
        knots0, knots1 = coll.dobj[wm][km]['knots']
        knots0 = coll.ddata[knots0]['data']
        knots1 = coll.ddata[knots1]['data']
        dx0 = np.min(np.diff(knots0))
        dx1 = np.min(np.diff(knots1))

    elif mtype == 'tri':
        indtri = coll.ddata[coll.dobj['mesh'][km]['ind']]['data']
        kknots = coll.dobj['mesh'][km]['knots']
        knots0 = coll.ddata[kknots[0]]['data']
        knots1 = coll.ddata[kknots[1]]['data']
        x0 = knots0[indtri]
        x1 = knots1[indtri]
        dist = np.mean(np.array([
            np.sqrt((x0[:, 1] - x0[:, 0])**2 + (x1[:, 1] - x1[:, 0])**2),
            np.sqrt((x0[:, 2] - x0[:, 1])**2 + (x1[:, 2] - x1[:, 1])**2),
            np.sqrt((x0[:, 2] - x0[:, 0])**2 + (x1[:, 2] - x1[:, 0])**2),
        ]))
        dx0, dx1 = dist, dist

    x0minmax = [knots0.min(), knots0.max()]
    x1minmax = [knots1.min(), knots1.max()]
    return dx0, dx1, x0minmax, x1minmax


# #############################################################################
# #############################################################################
#                       plot with submesh
# #############################################################################


def _plot_submesh(
    coll=None,
    collax=None,
    key=None,
    keym=None,
    keybs=None,
    # ref vetcor
    dref_vector=None,
    ref_vector_strategy=None,
    # plot_details
    plot_details=None,
    # figure
    vmin=None,
    vmax=None,
    cmap=None,
    color_dict=None,
    dax=None,
    dmargin=None,
    fs=None,
    dcolorbar=None,
    dleg=None,
    interp=None,
    dkeys=None,
    lcol=None,
):

    # ------------
    # check inputs

    plot_details = ds._generic_check._check_var(
        plot_details, 'plot_details',
        types=bool,
        default=False,
    )

    if dax is None:
        dax = _plot_profile2d_submesh_create_axes(
            fs=fs,
            dmargin=dmargin,
        )

    # ------------
    # preliminary

    # plot usual parts
    collax, dgroup = collax.plot_as_array(
        vmin=vmin,
        vmax=vmax,
        cmap=cmap,
        color_dict=color_dict,
        dax=dax,
        dmargin=dmargin,
        fs=fs,
        dcolorbar=dcolorbar,
        dleg=dleg,
        connect=False,
        interp=interp,
        label=True,
        inplace=True,
        **dkeys,
    )

    # ------------------
    # add radial profile to dax

    kradius, lkradial, lkdet, reft = _plot_profile2d_polar_add_radial(
        coll=coll,
        key=key,
        keym=keym,
        keybs=keybs,
        collax=collax,
        dref_vector=dref_vector,
        ref_vector_strategy=ref_vector_strategy,
        plot_details=plot_details,
    )

    if (reft is not None) != ('Z' in dgroup.keys()):
        msg = (
            "(reft is not None) != ('Z' in dgroup.keys()):\n"
            f"reft = {reft}\n"
            f"dgroup.keys() = {dgroup.keys()}\n"
            f"dref_vector = {dref_vector}"
        )
        raise Exception(msg)
    
    if reft is not None and reft not in dgroup['Z']['ref']:
        dgroup['Z']['ref'].append(reft)
        dgroup['Z']['data'].append('index')

    # ------------------
    # add radial profile

    kax = 'radial'
    lax = [k0 for k0, v0 in collax.dax.items() if kax in v0['type']]
    if len(lax) == 1:
        kax = lax[0]
        ax = collax.dax[kax]['handle']
        for ii in range(len(lkradial)):

            if reft is None:
                l0, = ax.plot(
                    collax.ddata[kradius]['data'],
                    collax.ddata[lkradial[ii]]['data'],
                    ls='-',
                    lw=2,
                    c=color_dict,
                )
            else:
                l0, = ax.plot(
                    collax.ddata[kradius]['data'],
                    collax.ddata[lkradial[ii]]['data'][0, :],
                    ls='-',
                    lw=2,
                    c=color_dict,
                )

                kl = f"{key}_radial{ii}"
                collax.add_mobile(
                    key=kl,
                    handle=l0,
                    refs=(reft,),
                    data=[lkradial[ii]],
                    dtype=['ydata'],
                    axes=kax,
                    ind=0,
                )

        if lkdet is not None:
            for ii in range(len(lkdet)):
                if reft is None:
                    l0, = ax.plot(
                        collax.ddata[kradius]['data'],
                        collax.ddata[lkdet[ii]]['data'],
                        ls='-',
                        lw=1,
                    )
                else:
                    l0, = ax.plot(
                        collax.ddata[kradius]['data'],
                        collax.ddata[lkdet[ii]]['data'][0, :],
                        ls='-',
                        lw=1,
                    )

                    kl = f"{key}_radial_det{ii}"
                    collax.add_mobile(
                        key=kl,
                        handle=l0,
                        refs=(reft,),
                        data=[lkdet[ii]],
                        dtype=['ydata'],
                        axes=kax,
                        ind=0,
                    )

        ax.set_xlim(
            collax.ddata[kradius]['data'].min(),
            collax.ddata[kradius]['data'].max(),
        )

        if vmin is not None:
            ax.set_ylim(bottom=vmin)
        if vmax is not None:
            ax.set_ylim(top=vmax)

    return collax, dgroup


def _plot_profile2d_polar_add_radial(
    coll=None,
    key=None,
    keym=None,
    keybs=None,
    collax=None,
    # ref_vector
    dref_vector=None,
    ref_vector_strategy=None,
    # details
    plot_details=None,
):

    # ----------
    # check

    if dref_vector is None:
        dref_vector = {}

    # -------------
    # key to radius

    wm = coll._which_mesh
    wbs = coll._which_bsplines

    kr2d = coll.dobj[wm][keym]['subkey'][0]
    kr = coll.dobj[wm][keym]['knots'][0]
    rr = coll.ddata[kr]['data']
    rad = np.linspace(rr[0], rr[-1], 10*rr.size)

    refr = f'{key}_nradius'
    kradius = f'{key}_radius'

    # -----------------
    # get angle if any

    clas = coll.dobj[wbs][keybs]['class']
    # if clas.knotsa is None:
    if True:
        angle = None
    elif len(clas.shapebs) == 2:
        ka = coll.dobj[wbs][keybs]['apex'][1]
        angle = coll.ddata[ka]['data']
    elif np.sum(clas.nbs_a_per_r > 1) == 1:
        i0 = (clas.nbs_a_per_r > 1).nonzero()[0][0]
        angle = coll.dobj[wbs][keybs]['class'].apex_per_bs_a[i0]
    else:
        pass

    # angle
    if angle is None:
        radmap = rad
        anglemap = angle
    else:
        radmap = np.repeat(rad[:, None], angle.size, axis=1)
        anglemap = np.repeat(angle[None, :], rad.size, axis=0)

    # ----
    # reft

    refc = [rr for rr in coll.ddata[key]['ref'] if rr in coll.ddata[kr2d]['ref']]
    if len(refc) == 1:
        refc = refc[0]
    else:
        refc = None

    # find reft
    reft, keyt, _, dind = coll.get_ref_vector_common(
        keys=[key, kr2d],
        ref=refc,
        strategy=ref_vector_strategy,
        **dref_vector,
    )[1:]

    # radial total profile
    # radial, t_radial, _ = coll.interpolate(
    dout = coll.interpolate(
        keys=key,
        ref_key=keybs,
        x0=radmap,
        x1=anglemap,
        grid=False,
        # ref_com=keyt,
    )[key]

    radial = dout['data']

    # if reft is not None and radial.ndim == radmap.ndim:
    #     radial = np.repeat(radial[None, ...], t_radial.size, axis=0)

    # -------------------------------
    # details for purely-radial cases

    if angle is None and plot_details is True:
        # radial_details, t_radial, _ = coll.interpolate(
        dout_details = coll.interpolate(
            ref_key=keybs,
            x0=rad,
            grid=False,
            details=True,
        )[f'{keybs}_details']

        radial_details = dout_details['data']

        if reft is None:
            radial_details = radial_details * coll.ddata[key]['data'][None, :]
            refdet = (refr,)
        else:
            refdet = (reft, refr)
            if reft == coll.get_time(key)[2]:
                radial_details = (
                    radial_details[None, :, :]
                    * coll.ddata[key]['data'][:, None, :]
                )
            elif key in dind.keys():
                radial_details = (
                    radial_details[None, :, :]
                    * coll.ddata[key]['data'][dind[key]['ind'], None, :]
                )

        nbs = radial_details.shape[-1]

    # -------------
    # add to collax

    collax.add_ref(key=refr, size=rad.size)
    if angle is not None:
        collax.add_ref(key='nangle', size=angle.size)

    ref = list(dout['ref'])
    ref[ref.index(None)] = refr
    ref = tuple(ref)

    # ------------
    # add to ddata

    lkdet = None
    collax.add_data(key=kradius, data=rad, ref=refr)
    if angle is None:
        lk = [f'{key}_radial']
        collax.add_data(key=lk[0], data=radial, ref=ref)
        if plot_details is True:
            lkdet = [f'{key}_radial-detail-{ii}' for ii in range(nbs)]
            for ii in range(nbs):
                collax.add_data(
                    key=lkdet[ii], data=radial_details[..., ii], ref=refdet,
                )

    else:
        kangle = 'angle'
        collax.add_data(key=kangle, data=angle, ref='nangle')
        lkdet = None
        lk = [f'radial-{ii}' for ii in range(angle.size)]
        if reft is None:
            for ii in range(angle.size):
                collax.add_data(key=lk[ii], data=radial[:, ii], ref=ref)
        else:
            for ii in range(angle.size):
                collax.add_data(key=lk[ii], data=radial[:, :, ii], ref=ref)

    return kradius, lk, lkdet, reft


# ##############################################################
# ##############################################################
#          create default axes for submesh
# ##############################################################


def _plot_profile2d_submesh_create_axes(
    fs=None,
    dmargin=None,
):

    if fs is None:
        fs = (15, 9)

    if dmargin is None:
        dmargin = {
            'left': 0.05, 'right': 0.95,
            'bottom': 0.05, 'top': 0.95,
            'hspace': 0.4, 'wspace': 0.3,
        }

    fig = plt.figure(figsize=fs)
    gs = gridspec.GridSpec(ncols=6, nrows=6, **dmargin)

    # axes for image
    ax0 = fig.add_subplot(gs[:4, 2:4], aspect='auto')

    # axes for vertical profile
    ax1 = fig.add_subplot(gs[:4, 4], sharey=ax0)

    # axes for horizontal profile
    ax2 = fig.add_subplot(gs[4:, 2:4], sharex=ax0)

    # axes for traces
    ax3 = fig.add_subplot(gs[2:4, :2])

    # axes for traces
    ax7 = fig.add_subplot(gs[:2, :2], sharey=ax2)

    # axes for text
    ax4 = fig.add_subplot(gs[:3, 5], frameon=False)
    ax5 = fig.add_subplot(gs[3:, 5], frameon=False)
    ax6 = fig.add_subplot(gs[4:, :2], frameon=False)

    # dax
    dax = {
        # data
        'matrix': {'handle': ax0},
        'vertical': {'handle': ax1},
        'horizontal': {'handle': ax2},
        'tracesZ': {'handle': ax3},
        'radial': {'handle': ax7},
        # text
        'textX': {'handle': ax4},
        'textY': {'handle': ax5},
        'textZ': {'handle': ax6},
    }
    return dax
